import json
from typing import Dict, List

from xopt.resources.testing import TEST_VOCS_BASE


class TestEnv:
    def test_basic_env(self):
        from badger.environment import Environment

        class Environment(Environment):
            name = "test"
            variables = {f"x{i}": [-1, 1] for i in range(20)}
            observables = ["f"]

            my_flag: int = 0

            def set_variables(self, variable_inputs: Dict[str, float]):
                pass

            def get_observables(self, observable_names: List[str]) -> Dict:
                return {ele: 1.0 for ele in observable_names}

        env = Environment()
        result = dict(env)
        assert result["my_flag"] == 0

        env.my_flag = 1
        result = dict(env)
        assert result["my_flag"] == 1

        result = json.loads(env.model_dump_json())
        assert result["my_flag"] == 1

    def test_env_in_routine(self):
        from badger.environment import Environment

        class Environment(Environment):
            name = "test"
            variables = {f"x{i}": [-1, 1] for i in range(20)}
            observables = ["f"]

            my_flag: int = 0

            def set_variables(self, variable_inputs: Dict[str, float]):
                pass

            def get_observables(self, observable_names: List[str]) -> Dict:
                return {ele: 1.0 for ele in observable_names}

        from badger.routine import Routine

        env = Environment()
        vocs = TEST_VOCS_BASE

        routine = Routine(
            name="test_routine", environment=env, vocs=vocs, generator="random"
        )
        result = json.loads(routine.json())
        assert result["environment"]["my_flag"] == 0

        routine.environment.my_flag = 1
        result = json.loads(routine.json())
        assert result["environment"]["my_flag"] == 1

    def test_env_from_get_env(self):
        from badger.factory import get_env
        from badger.routine import Routine

        env, config = get_env("test")
        config.pop("interface")
        vocs = TEST_VOCS_BASE

        routine = Routine(
            name="test_routine",
            environment=env(**config),
            vocs=vocs,
            generator="random",
        )
        result = json.loads(routine.json())
        assert result["environment"]["flag"] == 0

        routine.environment.flag = 1
        result = json.loads(routine.json())
        assert result["environment"]["flag"] == 1



from __future__ import print_function
import os
import json

from badger.logger.observer import _Tracker
from badger.logger.event import Events
from badger.logger.util import Colours


def _get_default_logger(verbose):
    return ScreenLogger(verbose=verbose)


class ScreenLogger(_Tracker):
    _default_cell_size = 10
    _default_precision = 4

    def __init__(self, verbose=2):
        self._verbose = verbose
        self._header_length = None
        super(ScreenLogger, self).__init__()

    @property
    def verbose(self):
        return self._verbose

    @verbose.setter
    def verbose(self, v):
        self._verbose = v

    def _format_number(self, x):
        if isinstance(x, int):
            s = "{x:< {s}}".format(
                x=x,
                s=self._default_cell_size,
            )
        else:
            s = "{x:< {s}.{p}}".format(
                x=x,
                s=self._default_cell_size,
                p=self._default_precision,
            )

        if len(s) > self._default_cell_size:
            if "." in s:
                return s[: self._default_cell_size]
            else:
                return s[: self._default_cell_size - 3] + "..."
        return s

    def _format_key(self, key):
        s = "{key:^{s}}".format(key=key, s=self._default_cell_size)
        if len(s) > self._default_cell_size:
            return s[: self._default_cell_size - 3] + "..."
        return s

    def _step(self, solution, colour=Colours.black):
        # solution: (x: 1d array, y: 1d array, c: 1d array, s: 1d array, is_optimal: bool,
        #            vars: str list, obses: str list, cons: str list, stas: str list)
        cells = []

        cells.append(self._format_number(self._iterations + 1))

        for o in solution[1]:
            cells.append(self._format_number(o))

        for c in solution[2]:
            cells.append(self._format_number(c))

        for v in solution[0]:
            cells.append(self._format_number(v))

        for s in solution[3]:
            cells.append(self._format_number(s))  # TODO: deal with the str case

        return "| " + " | ".join(map(colour, cells)) + " |"

    def _header(self, solution):
        cells = []
        cells.append(self._format_key("iter"))

        for obs in solution[6]:
            cells.append(self._format_key(obs))

        for con in solution[7]:
            cells.append(self._format_key(con))

        for var in solution[5]:
            cells.append(self._format_key(var))

        for sta in solution[8]:
            cells.append(self._format_key(sta))

        line = "| " + " | ".join(cells) + " |"
        self._header_length = len(line)
        return line + "\n" + ("-" * self._header_length)

    def _is_new_max(self, solution):
        return solution[4]

    def update(self, event, solution):
        if event == Events.OPTIMIZATION_START:
            line = self._header(solution) + "\n"
        elif event == Events.OPTIMIZATION_STEP:
            is_new_max = self._is_new_max(solution)
            if self._verbose == 1 and not is_new_max:
                line = ""
            else:
                colour = Colours.purple if is_new_max else Colours.black
                line = self._step(solution, colour=colour) + "\n"
        elif event == Events.OPTIMIZATION_END:
            line = "=" * self._header_length + "\n"

        if self._verbose:
            print(line, end="")
        self._update_tracker(event, solution)


class JSONLogger(_Tracker):
    def __init__(self, path, reset=True):
        self._path = path if path[-5:] == ".json" else path + ".json"
        if reset:
            try:
                os.remove(self._path)
            except OSError:
                pass
        super(JSONLogger, self).__init__()

    def update(self, event, solution):
        if event == Events.OPTIMIZATION_STEP:
            data = {
                "x": solution[0],
                "y": solution[1],
                "c": solution[2],
                "s": solution[3],
                "is_optimal": solution[4],
            }

            now, time_elapsed, time_delta = self._time_metrics()
            data["datetime"] = {
                "datetime": now,
                "elapsed": time_elapsed,
                "delta": time_delta,
            }

            with open(self._path, "a") as f:
                f.write(json.dumps(data) + "\n")

        self._update_tracker(event, solution)



import os
import warnings
from datetime import datetime
import logging

import yaml
import sqlite3
import uuid

from badger.routine import Routine
from badger.settings import init_settings
from badger.utils import get_yaml_string
from badger.errors import BadgerConfigError, BadgerDBError

logger = logging.getLogger(__name__)

# Check badger database root
flag_use_db = True
config_singleton = init_settings()
try:
    BADGER_DB_ROOT = config_singleton.read_value("BADGER_DB_ROOT")
except KeyError:
    flag_use_db = False

if flag_use_db:
    if BADGER_DB_ROOT is None:
        raise BadgerConfigError("Please set the BADGER_DB_ROOT env var!")
    elif not os.path.exists(BADGER_DB_ROOT):
        os.makedirs(BADGER_DB_ROOT)
        logger.info(f"Badger database root {BADGER_DB_ROOT} created")


def ensure_routines_db_exists(func):
    """
    Create the routines database (a SQL table) if it does not already exist.
    """

    def func_safe(*args, **kwargs):
        db_routine = os.path.join(BADGER_DB_ROOT, "routines.db")

        con = sqlite3.connect(db_routine)
        cur = con.cursor()

        cur.execute(
            "create table if not exists routine (id text primary key, name text, config, savedAt timestamp)"
        )

        con.commit()
        con.close()

        return func(*args, **kwargs)

    return func_safe


def ensure_runs_db_exists(func):
    """
    Create the runs database (a SQL table) if it does not already exist.
    """

    def func_safe(*args, **kwargs):
        db_run = os.path.join(BADGER_DB_ROOT, "runs.db")

        con = sqlite3.connect(db_run)
        cur = con.cursor()

        cur.execute(
            "create table if not exists run (id integer primary key, savedAt timestamp, finishedAt timestamp, routine_id, filename)"
        )

        con.commit()
        con.close()

        return func(*args, **kwargs)

    return func_safe


def filter_routines(records, tags):
    records_filtered = []
    for record in records:
        try:
            _tags = yaml.safe_load(record[3])["config"]["tags"]
            if tags.items() <= _tags.items():
                records_filtered.append(record)
        except:
            pass

    return records_filtered


def extract_metadata(records):
    env_list = []
    descr_list = []
    for record in records:
        try:
            metadata = yaml.safe_load(record[2])
            env = metadata["environment"]["name"]
            env_list.append(env)
            descr = metadata["description"]
            descr_list.append(descr)
        except Exception:
            env_list.append("")
            descr_list.append("")

    return env_list, descr_list


@ensure_routines_db_exists
def save_routine(routine: Routine):
    db_routine = os.path.join(BADGER_DB_ROOT, "routines.db")

    con = sqlite3.connect(db_routine)
    cur = con.cursor()

    id = str(uuid.uuid4())
    routine.id = id
    cur.execute(
        "insert into routine values (?, ?, ?, ?)",
        (routine.id, routine.name, routine.yaml(), datetime.now()),
    )

    con.commit()
    con.close()


# This function is not safe and might break database! Use with caution!
@ensure_routines_db_exists
def update_routine(routine: Routine):
    db_routine = os.path.join(BADGER_DB_ROOT, "routines.db")

    con = sqlite3.connect(db_routine)
    cur = con.cursor()

    cur.execute("select * from routine where id=:id", {"id": routine.id})
    record = cur.fetchone()

    if record:  # update the record
        cur.execute(
            "update routine set name = ?, config = ?, savedAt = ? where id = ?",
            (routine.name, routine.yaml(), datetime.now(), routine.id),
        )

    con.commit()
    con.close()


@ensure_routines_db_exists
@ensure_runs_db_exists
def remove_routine(id: str, remove_runs=True):
    db_routine = os.path.join(BADGER_DB_ROOT, "routines.db")

    con = sqlite3.connect(db_routine)
    cur = con.cursor()

    cur.execute(f'delete from routine where id = "{id}"')

    con.commit()
    con.close()

    if remove_runs:
        # Remove all related run records
        db_run = os.path.join(BADGER_DB_ROOT, "runs.db")

        con = sqlite3.connect(db_run)
        cur = con.cursor()

        cur.execute(f'delete from run where routine_id = "{id}"')

        con.commit()
        con.close()


@ensure_routines_db_exists
def load_routine(id: str):
    db_routine = os.path.join(BADGER_DB_ROOT, "routines.db")
    con = sqlite3.connect(db_routine)
    cur = con.cursor()

    if isinstance(id, str) and id.strip():
        cur.execute("select * from routine where id=:id", {"id": id})
    else:
        raise ValueError("Expected id to be a non-empty string.")

    records = cur.fetchall()
    con.close()

    if len(records) == 1:
        # return yaml.safe_load(records[0][1]), records[0][2]
        routine_dict = yaml.safe_load(records[0][2])
        # routine_dict['evaluator'] = None
        with warnings.catch_warnings(record=True) as caught_warnings:
            routine = Routine(**routine_dict)

            # Check if any user warnings were caught
            for warning in caught_warnings:
                if isinstance(warning.category, UserWarning):
                    pass
                else:
                    print(f"Caught user warning: {warning.message}")

            return routine, records[0][3]
    elif len(records) == 0:
        raise BadgerDBError(f"Routine id {id} not found in the database!")
    else:
        raise BadgerDBError(f"Multiple routines with id {id} found in the database!")


@ensure_routines_db_exists
def list_routine(keyword="", tags={}):
    db_routine = os.path.join(BADGER_DB_ROOT, "routines.db")
    con = sqlite3.connect(db_routine)
    cur = con.cursor()

    # check if id column is in database
    # if not, add it and update routine and run entries accordingly
    cur.execute("pragma table_info(routine)")
    columns = [row[1] for row in cur.fetchall()]
    if "id" not in columns:
        cur.execute(
            """
        create table new_table (
            id text primary key,
            name text,
            config,
            savedAt timestamp
        )
        """
        )
        db_run = os.path.join(BADGER_DB_ROOT, "runs.db")
        con_run = sqlite3.connect(db_run)
        cur_run = con_run.cursor()
        cur_run.execute("alter table run rename column routine to routine_id")
        con_run.commit()
        con_run.close()
        cur.execute("select * from routine")
        rows = cur.fetchall()
        for row in rows:
            id = str(uuid.uuid4())
            config = yaml.safe_load(row[1])
            config["id"] = id
            sorted_config = dict(sorted(config.items()))
            new_config = yaml.dump(sorted_config, default_flow_style=False)
            cur.execute(
                "insert into new_table (id, name, config, savedAt) values (?, ?, ?, ?)",
                (id, row[0], new_config, row[2]),
            )
            db_run = os.path.join(BADGER_DB_ROOT, "runs.db")
            con_run = sqlite3.connect(db_run)
            cur_run = con_run.cursor()
            # now the column in run table is called 'routine_id' but they are still routine names
            cur_run.execute(
                "update run set routine_id = ? where routine_id = ?", (id, row[0])
            )
            con_run.commit()
            con_run.close()
            filenames = get_runs_by_routine(id)
            # Check badger optimization run archive root
            config_singleton = init_settings()
            BADGER_ARCHIVE_ROOT = config_singleton.read_value("BADGER_ARCHIVE_ROOT")
            if BADGER_ARCHIVE_ROOT is None:
                raise BadgerConfigError("Please set the BADGER_ARCHIVE_ROOT env var!")
            elif not os.path.exists(BADGER_ARCHIVE_ROOT):
                os.makedirs(BADGER_ARCHIVE_ROOT)
                logger.info(f"Badger run root {BADGER_ARCHIVE_ROOT} created")
            for i, fname in enumerate(filenames):
                tokens = fname.split("-")
                first_level = tokens[1]
                second_level = f"{tokens[1]}-{tokens[2]}"
                third_level = f"{tokens[1]}-{tokens[2]}-{tokens[3]}"

                filename = os.path.join(
                    BADGER_ARCHIVE_ROOT, first_level, second_level, third_level, fname
                )
                filenames[i] = filename
            for filename in filenames:
                with open(filename, "r") as file:
                    run = yaml.safe_load(file)
                run["id"] = id
                sorted_run = {key: run[key] for key in sorted(run.keys())}
                with open(filename, "w") as file:
                    yaml.dump(sorted_run, file, default_flow_style=False)
        cur.execute("drop table routine")
        cur.execute("alter table new_table rename to routine")
        con.commit()

    cur.execute(
        f"select id, name, config, savedAt from routine where name like '%{keyword}%' order by savedAt desc"
    )
    records = cur.fetchall()
    if tags:
        records = filter_routines(records, tags)
    ids = [record[0] for record in records]
    names = [record[1] for record in records]
    timestamps = [record[3] for record in records]
    environments, descriptions = extract_metadata(records)
    con.close()

    return ids, names, timestamps, environments, descriptions


@ensure_runs_db_exists
def save_run(run):
    db_run = os.path.join(BADGER_DB_ROOT, "runs.db")

    con = sqlite3.connect(db_run, timeout=30.0)
    cur = con.cursor()

    # Insert or update a record
    routine_id = run["routine"].id
    run_filename = run["filename"]
    timestamps = run["data"]["timestamp"]
    time_start = datetime.fromtimestamp(timestamps[0])
    time_finish = datetime.fromtimestamp(timestamps[-1])

    # Check if the record exist (same filename)
    cur.execute("select id from run where filename = ?", (run_filename,))
    existing_row = cur.fetchone()

    if existing_row:
        cur.execute(
            "update run set finishedAt = ? where filename = ?",
            (time_finish, run_filename),
        )
        rid = existing_row[0]
    else:
        cur.execute(
            "insert into run values (?, ?, ?, ?, ?)",
            (None, time_start, time_finish, routine_id, run_filename),
        )
        rid = cur.lastrowid

    con.commit()
    con.close()

    return rid


@ensure_runs_db_exists
def get_runs_by_routine(routine_id: str):
    db_run = os.path.join(BADGER_DB_ROOT, "runs.db")

    con = sqlite3.connect(db_run)
    cur = con.cursor()

    cur.execute(
        f'select filename from run where routine_id = "{routine_id}" order by savedAt desc'
    )
    records = cur.fetchall()
    con.close()

    filenames = [record[0] for record in records]

    return filenames


@ensure_runs_db_exists
def get_runs():
    db_run = os.path.join(BADGER_DB_ROOT, "runs.db")

    con = sqlite3.connect(db_run)
    cur = con.cursor()

    cur.execute("select filename from run order by savedAt desc")
    records = cur.fetchall()

    con.close()

    filenames = [record[0] for record in records]

    return filenames


@ensure_runs_db_exists
def remove_run_by_filename(filename):
    db_run = os.path.join(BADGER_DB_ROOT, "runs.db")

    con = sqlite3.connect(db_run)
    cur = con.cursor()

    cur.execute(f'delete from run where filename = "{filename}"')

    con.commit()
    con.close()


@ensure_runs_db_exists
def remove_run_by_id(rid):
    db_run = os.path.join(BADGER_DB_ROOT, "runs.db")

    con = sqlite3.connect(db_run)
    cur = con.cursor()

    cur.execute(f"delete from run where id = {rid}")

    con.commit()
    con.close()


def import_routines(filename):
    con = sqlite3.connect(filename)
    cur = con.cursor()

    # Deal with empty db file
    cur.execute(
        "create table if not exists routine (id text primary key, name text, config, savedAt timestamp)"
    )

    db_routine = os.path.join(BADGER_DB_ROOT, "routines.db")
    con_db = sqlite3.connect(db_routine)
    cur_db = con_db.cursor()

    cur.execute("select * from routine")
    records = cur.fetchall()

    failed_list = []
    for record in records:
        try:
            cur_db.execute("insert into routine values (?, ?, ?, ?)", record)
        except:
            failed_list.append(record[0])

    con_db.commit()
    con_db.close()

    con.close()

    if failed_list:
        raise BadgerDBError(get_yaml_string(failed_list))


def export_routines(filename, routine_id_list):
    con = sqlite3.connect(filename)
    cur = con.cursor()

    cur.execute(
        "create table if not exists routine (id text primary key, name text, config, savedAt timestamp)"
    )

    db_routine = os.path.join(BADGER_DB_ROOT, "routines.db")
    con_db = sqlite3.connect(db_routine)
    cur_db = con_db.cursor()

    for id in routine_id_list:
        cur_db.execute("select * from routine where id=:id", {"id": id})
        records = cur_db.fetchall()
        record = records[0]  # should only have one hit

        cur.execute("insert into routine values (?, ?, ?, ?)", record)

    con_db.close()

    con.commit()
    con.close()



def test_find_intf():
    from badger.factory import get_intf, list_intf

    assert len(list_intf()) == 1

    _, configs = get_intf("test")
    assert configs["name"] == "test"
    assert configs["version"] == "1.0"


def test_get_params():
    from badger.factory import get_intf

    _, configs = get_intf("test")
    params = configs["params"]

    assert params == {"flag": 0}


def test_set_params():
    from badger.factory import get_intf

    Interface, _ = get_intf("test")
    intf = Interface(flag=1)

    assert intf.flag == 1


def test_get_values():
    from badger.factory import get_intf

    Interface, _ = get_intf("test")
    intf = Interface()

    channel_outputs = intf.get_values(["x1", "x2"])
    assert channel_outputs == {"x1": 0.5, "x2": 0.5}

    # Test single version
    value = intf.get_value("x3")
    assert value == 0.5


def test_set_values():
    from badger.factory import get_intf

    Interface, _ = get_intf("test")
    intf = Interface()

    channel_inputs = {"x1": 3, "x2": 4}
    intf.set_values(channel_inputs)
    channel_outputs = intf.get_values(["x1", "x2", "x3"])
    assert channel_outputs == {"x1": 3, "x2": 4, "x3": 0.5}

    # Test single version
    intf.set_value("x3", 5)
    value = intf.get_value("x3")
    assert value == 5


def test_recording():
    from badger.factory import get_intf

    Interface, _ = get_intf("test")
    intf = Interface()

    intf.set_values({"x1": 3, "x2": 4})
    intf.set_value("x3", 5)
    intf.get_values(["x1", "x2", "x3"])

    assert len(intf._logs) == 3
    record = intf._logs[0]
    assert record["action"] == "set_values"
    assert record["channel_inputs"] == {"x1": 3, "x2": 4}
    record = intf._logs[1]
    assert record["action"] == "set_values"
    assert record["channel_inputs"] == {"x3": 5}
    record = intf._logs[2]
    assert record["action"] == "get_values"
    assert record["channel_outputs"] == {"x1": 3, "x2": 4, "x3": 5}


# def test_run(mock_config_root):
#     from coolname import generate_slug
#     from badger.log import config_log
#     config_log()
#     from badger.factory import get_generator, get_env
#     from badger.utils import load_config, merge_params, config_list_to_dict
#     from badger.core import normalize_routine, run_routine

#     # Args
#     generator = 'silly'
#     generator_params = '{dimension: 2, max_iter: 10}'
#     env = 'silly'
#     env_params = None
#     config = os.path.join(mock_config_root, 'test.yaml')
#     save = False
#     yes = True
#     verbose = 2

#     # Get env params
#     _, configs_env = get_env(env)

#     # Get generator params
#     _, configs_generator = get_generator(generator)

#     # Normalize the generator and env params
#     params_env = load_config(env_params)
#     params_generator = load_config(generator_params)
#     params_env = merge_params(configs_env['params'], params_env)
#     params_generator = merge_params(configs_generator['params'], params_generator)

#     # Load routine configs
#     configs_routine = load_config(config)

#     # Compose the routine
#     routine = {
#         'name': save or generate_slug(2),
#         'generator': generator,
#         'env': env,
#         'generator_params': params_generator,
#         'env_params': params_env,
#         # env_vranges is an additional info for the normalization
#         # Will be removed after the normalization
#         'env_vranges': config_list_to_dict(configs_env['variables']),
#         'config': configs_routine,
#     }

#     # Sanity check and config normalization
#     routine = normalize_routine(routine)

#     run_routine(routine, yes, save, verbose)



from importlib import resources

from PyQt5.QtWidgets import (
    QVBoxLayout,
    QHBoxLayout,
    QPushButton,
    QWidget,
    QPlainTextEdit,
    QLineEdit,
)
from PyQt5.QtWidgets import (
    QCheckBox,
    QStyledItemDelegate,
    QLabel,
    QListWidget,
    QSizePolicy,
)
from PyQt5.QtGui import QIcon, QFont
from PyQt5.QtCore import QRegExp, QPropertyAnimation

from badger.gui.default.components.collapsible_box import CollapsibleBox
from badger.gui.default.components.var_table import VariableTable
from badger.gui.default.components.obj_table import ObjectiveTable
from badger.gui.default.components.con_table import ConstraintTable
from badger.gui.default.components.data_table import init_data_table
from badger.settings import init_settings
from badger.gui.default.utils import (
    MouseWheelWidgetAdjustmentGuard,
    NoHoverFocusComboBox,
)
from badger.utils import strtobool

LABEL_WIDTH = 96
ENV_PARAMS_BTN = 1  # use button or collapsible box for env parameters

stylesheet_auto = """
    #VarPanel {
        border: 4px solid #FDD835;
        border-radius: 4px;
    }
"""

stylesheet_manual = """
    #VarPanel {
        border: 4px solid #60798B;
        border-radius: 4px;
    }
"""

stylesheet_auto_msg = """
    QLabel {
        background-color: #FDD835;
        color: #19232D;
        padding: 4px 4px 8px 4px;
        border-radius: 0;
    }
"""

stylesheet_manual_msg = """
    QLabel {
        background-color: #60798B;
        color: #FFFFFF;
        padding: 4px 4px 8px 4px;
        border-radius: 0;
    }
"""

stylesheet_load = """
QPushButton:hover:pressed
{
    background-color: #4DB6AC;
}
QPushButton:hover
{
    background-color: #26A69A;
}
QPushButton
{
    color: #FFFFFF;
    background-color: #00897B;
}
"""

MSG_AUTO_RELATIVE = (
    "The values you see in the variable ranges spin "
    "boxes and initial points table are preview that "
    "generated either based on current machine state, "
    "or the machine state at the time of this historical"
    " run. The real values would be regenerated based on"
    " the machine state before running the optimization "
    "again."
)

MSG_AUTO = (
    "Auto mode is on.  To manually set the "
    "variable ranges and/or initial points,  please "
    'uncheck the "Automatic" check box.'
)

MSG_MANUAL = (
    "Auto mode is off.  To automatically set the "
    "variable ranges and/or initial points,  please "
    'check the "Automatic" check box.'
)


class BadgerEnvBox(QWidget):
    def __init__(self, env_dict, parent=None, envs=[]):
        super().__init__(parent)

        self.envs = envs
        self.env_dict = env_dict

        self.init_ui()
        self.config_logic()

    def init_ui(self):
        config_singleton = init_settings()

        icon_ref = resources.files(__package__) / "../images/import.png"
        with resources.as_file(icon_ref) as icon_path:
            self.icon_import = QIcon(str(icon_path))

        vbox = QVBoxLayout(self)
        vbox.setContentsMargins(8, 8, 8, 8)

        # Load Template Button
        template_button = QWidget()
        # template_button.setFixedWidth(128)
        hbox_temp = QHBoxLayout(template_button)
        hbox_temp.setContentsMargins(0, 0, 0, 0)

        cool_font = QFont()
        cool_font.setWeight(QFont.DemiBold)

        self.load_template_button = load_template_button = QPushButton(" Load Template")
        load_template_button.setFixedHeight(36)
        load_template_button.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        load_template_button.setIcon(self.icon_import)
        load_template_button.setFont(cool_font)
        load_template_button.setStyleSheet(stylesheet_load)

        hbox_temp.addWidget(load_template_button, 1)
        vbox.addWidget(template_button)
        template_button.show()

        self.setObjectName("EnvBox")

        name = QWidget()
        hbox_name = QHBoxLayout(name)
        hbox_name.setContentsMargins(0, 0, 0, 0)
        lbl = QLabel("Environment")
        lbl.setFixedWidth(LABEL_WIDTH)
        self.cb = cb = NoHoverFocusComboBox()
        cb.setItemDelegate(QStyledItemDelegate())
        cb.addItems(self.envs)
        cb.setCurrentIndex(-1)
        cb.installEventFilter(MouseWheelWidgetAdjustmentGuard(cb))
        self.btn_env_play = btn_env_play = QPushButton("Open Playground")
        btn_env_play.setFixedSize(128, 24)
        if not strtobool(config_singleton.read_value("BADGER_ENABLE_ADVANCED")):
            btn_env_play.hide()
        self.btn_pv = btn_pv = QPushButton("Variable Search")
        btn_pv.setFixedSize(128, 24)
        self.btn_docs = btn_docs = QPushButton("Open Docs")
        btn_docs.setFixedSize(96, 24)
        self.btn_params = btn_params = QPushButton("Parameters")
        btn_params.setCheckable(True)
        btn_params.setFixedSize(96, 24)
        if not ENV_PARAMS_BTN:
            btn_params.hide()
        hbox_name.addWidget(lbl)
        hbox_name.addWidget(cb, 1)
        hbox_name.addWidget(btn_env_play)
        hbox_name.addWidget(btn_params)
        hbox_name.addWidget(btn_pv)
        hbox_name.addWidget(btn_docs)
        vbox.addWidget(name)

        self.edit = edit = QPlainTextEdit()
        # edit.setMinimumHeight(480)
        if ENV_PARAMS_BTN:
            vbox.addWidget(edit)
            edit.setMaximumHeight(0)
            edit.hide()
        else:
            params = QWidget()
            vbox.addWidget(params)
            hbox_params = QHBoxLayout(params)
            hbox_params.setContentsMargins(0, 0, 0, 0)
            lbl_params_col = QWidget()
            vbox_lbl_params = QVBoxLayout(lbl_params_col)
            vbox_lbl_params.setContentsMargins(0, 0, 0, 0)
            lbl_params = QLabel("")
            lbl_params.setFixedWidth(LABEL_WIDTH)
            vbox_lbl_params.addWidget(lbl_params)
            vbox_lbl_params.addStretch(1)
            hbox_params.addWidget(lbl_params_col)

            edit_params_col = QWidget()
            hbox_params.addWidget(edit_params_col)
            vbox_params_edit = QVBoxLayout(edit_params_col)
            vbox_params_edit.setContentsMargins(0, 0, 0, 0)

            cbox_params = CollapsibleBox(self, " Parameters")
            vbox_params_edit.addWidget(cbox_params, 1)
            vbox_params = QVBoxLayout()

            vbox_params.addWidget(edit, 1)
            cbox_params.setContentLayout(vbox_params)

        self.animation = QPropertyAnimation(self.edit, b"maximumHeight")
        self.animation.setDuration(150)

        # vbox.addWidget(edit)
        # vbox_params_edit.addWidget(edit)
        # hbox_params.addWidget(edit_params_col)
        # vbox.addWidget(params)

        # seperator = QFrame()
        # seperator.setFrameShape(QFrame.HLine)
        # seperator.setFrameShadow(QFrame.Sunken)
        # seperator.setLineWidth(0)
        # seperator.setMidLineWidth(0)
        # vbox.addWidget(seperator)

        # Variables config (table style)
        self.var_panel = var_panel = QWidget()
        var_panel.setObjectName("VarPanel")
        var_panel.setMinimumHeight(280)
        vbox.addWidget(var_panel, 2)
        self.vbox_var = vbox_var = QVBoxLayout(var_panel)
        vbox_var.setContentsMargins(4, 4, 4, 4)
        vbox_var.setSpacing(0)
        if config_singleton.read_value("AUTO_REFRESH"):
            self.MSG_AUTO = MSG_AUTO_RELATIVE
            self.MSG_MANUAL = MSG_MANUAL
        else:
            self.MSG_AUTO = MSG_AUTO
            self.MSG_MANUAL = MSG_MANUAL
        msg_auto = QLabel(self.MSG_AUTO)
        msg_auto.setWordWrap(True)
        msg_auto.setStyleSheet(stylesheet_auto_msg)
        self.msg_auto = msg_auto
        vbox_var.addWidget(msg_auto)
        var_panel_origin = QWidget()
        vbox_var.addWidget(var_panel_origin)
        self.hbox_var = hbox_var = QHBoxLayout(var_panel_origin)
        hbox_var.setContentsMargins(8, 8, 8, 8)
        lbl_var_col = QWidget()
        vbox_lbl_var = QVBoxLayout(lbl_var_col)
        vbox_lbl_var.setContentsMargins(0, 0, 0, 0)
        lbl_var = QLabel("Variables")
        lbl_var.setFixedWidth(LABEL_WIDTH)
        vbox_lbl_var.addWidget(lbl_var)
        vbox_lbl_var.addStretch(1)
        hbox_var.addWidget(lbl_var_col)

        edit_var_col = QWidget()
        vbox_var_edit = QVBoxLayout(edit_var_col)
        vbox_var_edit.setContentsMargins(0, 0, 0, 0)

        # Add relative to current option
        action_common = QWidget()
        hbox_action_common = QHBoxLayout(action_common)
        hbox_action_common.setContentsMargins(0, 0, 0, 0)
        vbox_var_edit.addWidget(action_common)
        self.relative_to_curr = relative_to_curr = QCheckBox("Automatic")
        relative_to_curr.setChecked(False)
        tooltip = (
            "If checked, you will not be able to change the\n"
            + "variable ranges and initial points manually.\n"
            + "Instead, the variable ranges and the initial points will be\n"
            + "generated based on the current state.\n\n"
            + 'You can adjust them by using the "Set Variable Range"\n'
            + 'button and the "Add Current"/"Add Random" buttons.\n'
            + "The actual values of those settings will be re-calculated\n"
            + "based on the machine state at the time of running."
        )
        relative_to_curr.setToolTip(tooltip)
        hbox_action_common.addWidget(relative_to_curr)
        self.btn_refresh = btn_refresh = QPushButton("Refresh")
        btn_refresh.setFixedSize(96, 24)
        btn_refresh.setDisabled(True)
        tooltip = (
            "Refresh the variable ranges and the initial points based on\n"
            + "the current variable values.\n\n"
            + 'Note that in manual mode, click "Refresh" will clear the \n'
            + "initial points table if the variable ranges change,\n"
            + "since the old initial points might be invalid.\n"
            + "In this case, you will need to add initial points again."
        )
        btn_refresh.setToolTip(tooltip)
        hbox_action_common.addWidget(btn_refresh)
        hbox_action_common.addStretch()

        action_var = QWidget()
        hbox_action_var = QHBoxLayout(action_var)
        hbox_action_var.setContentsMargins(0, 0, 0, 0)
        vbox_var_edit.addWidget(action_var)
        self.edit_var = edit_var = QLineEdit()
        edit_var.setPlaceholderText("Filter variables...")
        edit_var.setFixedWidth(160)
        self.btn_add_var = btn_add_var = QPushButton("Add")
        btn_add_var.setFixedSize(96, 24)
        btn_add_var.setDisabled(True)
        if not strtobool(config_singleton.read_value("BADGER_ENABLE_ADVANCED")):
            btn_add_var.hide()
        self.btn_lim_vrange = btn_lim_vrange = QPushButton("Set Variable Range")
        btn_lim_vrange.setFixedSize(144, 24)
        btn_lim_vrange.setDisabled(True)
        self.check_only_var = check_only_var = QCheckBox("Show Checked Only")
        check_only_var.setChecked(False)
        hbox_action_var.addWidget(edit_var)
        hbox_action_var.addWidget(btn_add_var)
        hbox_action_var.addWidget(btn_lim_vrange)
        hbox_action_var.addStretch()
        hbox_action_var.addWidget(check_only_var)

        self.var_table = VariableTable()
        self.var_table.lock_bounds()
        vbox_var_edit.addWidget(self.var_table)

        cbox_init = CollapsibleBox(
            self,
            " Initial Points",
            tooltip="If set, it takes precedence over the start from current setting in generator configuration.",
        )
        vbox_var_edit.addWidget(cbox_init)
        vbox_init = QVBoxLayout()
        vbox_init.setContentsMargins(8, 8, 8, 8)
        action_init = QWidget()
        hbox_action_init = QHBoxLayout(action_init)
        hbox_action_init.setContentsMargins(0, 0, 0, 0)
        self.btn_add_row = btn_add_row = QPushButton("Add Row")
        btn_add_row.setFixedSize(96, 24)
        self.btn_add_curr = btn_add_curr = QPushButton("Add Current")
        btn_add_curr.setFixedSize(96, 24)
        self.btn_add_rand = btn_add_rand = QPushButton("Add Random")
        btn_add_rand.setFixedSize(96, 24)
        self.btn_clear = btn_clear = QPushButton("Clear All")
        btn_clear.setFixedSize(96, 24)
        hbox_action_init.addWidget(btn_add_row)
        hbox_action_init.addStretch()
        hbox_action_init.addWidget(btn_add_curr)
        hbox_action_init.addWidget(btn_add_rand)
        hbox_action_init.addWidget(btn_clear)
        vbox_init.addWidget(action_init)
        self.init_table = init_data_table()
        self.init_table.set_uneditable()
        vbox_init.addWidget(self.init_table)
        cbox_init.setContentLayout(vbox_init)
        cbox_init.expand()

        hbox_var.addWidget(edit_var_col)

        # Objectives config (table style)
        obj_panel = QWidget()
        vbox.addWidget(obj_panel, 1)
        hbox_obj = QHBoxLayout(obj_panel)
        hbox_obj.setContentsMargins(0, 0, 0, 0)
        lbl_obj_col = QWidget()
        vbox_lbl_obj = QVBoxLayout(lbl_obj_col)
        vbox_lbl_obj.setContentsMargins(0, 0, 0, 0)
        lbl_obj = QLabel("Objectives")
        lbl_obj.setFixedWidth(LABEL_WIDTH)
        vbox_lbl_obj.addWidget(lbl_obj)
        vbox_lbl_obj.addStretch(1)
        hbox_obj.addWidget(lbl_obj_col)

        edit_obj_col = QWidget()
        vbox_obj_edit = QVBoxLayout(edit_obj_col)
        vbox_obj_edit.setContentsMargins(0, 0, 0, 0)

        action_obj = QWidget()
        hbox_action_obj = QHBoxLayout(action_obj)
        hbox_action_obj.setContentsMargins(0, 0, 0, 0)
        vbox_obj_edit.addWidget(action_obj)
        self.edit_obj = edit_obj = QLineEdit()
        edit_obj.setPlaceholderText("Filter objectives...")
        edit_obj.setFixedWidth(192)
        self.check_only_obj = check_only_obj = QCheckBox("Show Checked Only")
        check_only_obj.setChecked(False)
        hbox_action_obj.addWidget(edit_obj)
        hbox_action_obj.addStretch()
        hbox_action_obj.addWidget(check_only_obj)

        self.obj_table = ObjectiveTable()
        vbox_obj_edit.addWidget(self.obj_table)
        hbox_obj.addWidget(edit_obj_col)

        cbox_more = CollapsibleBox(self, " More")
        vbox.addWidget(cbox_more)
        vbox_more = QVBoxLayout()

        # Constraints config
        con_panel = QWidget()
        vbox_more.addWidget(con_panel, 1)
        hbox_con = QHBoxLayout(con_panel)
        hbox_con.setContentsMargins(0, 0, 0, 0)
        lbl_con_col = QWidget()
        vbox_lbl_con = QVBoxLayout(lbl_con_col)
        vbox_lbl_con.setContentsMargins(0, 0, 0, 0)
        lbl_con = QLabel("Constraints")
        lbl_con.setFixedWidth(LABEL_WIDTH)
        vbox_lbl_con.addWidget(lbl_con)
        vbox_lbl_con.addStretch(1)
        hbox_con.addWidget(lbl_con_col)

        edit_con_col = QWidget()
        vbox_con_edit = QVBoxLayout(edit_con_col)
        vbox_con_edit.setContentsMargins(0, 0, 0, 0)

        action_con = QWidget()
        hbox_action_con = QHBoxLayout(action_con)
        hbox_action_con.setContentsMargins(0, 0, 0, 0)
        vbox_con_edit.addWidget(action_con)
        self.btn_add_con = btn_add_con = QPushButton("Add")
        btn_add_con.setFixedSize(96, 24)
        btn_add_con.setDisabled(True)
        hbox_action_con.addWidget(btn_add_con)
        hbox_action_con.addStretch()

        self.con_table = ConstraintTable()
        self.con_table.setMinimumHeight(120)
        vbox_con_edit.addWidget(self.con_table)
        hbox_con.addWidget(edit_con_col)

        # States config
        sta_panel = QWidget()
        vbox_more.addWidget(sta_panel, 1)
        hbox_sta = QHBoxLayout(sta_panel)
        hbox_sta.setContentsMargins(0, 0, 0, 0)
        lbl_sta_col = QWidget()
        vbox_lbl_sta = QVBoxLayout(lbl_sta_col)
        vbox_lbl_sta.setContentsMargins(0, 0, 0, 0)
        lbl_sta = QLabel("Observables")
        lbl_sta.setFixedWidth(LABEL_WIDTH)
        vbox_lbl_sta.addWidget(lbl_sta)
        vbox_lbl_sta.addStretch(1)
        hbox_sta.addWidget(lbl_sta_col)

        edit_sta_col = QWidget()
        vbox_sta_edit = QVBoxLayout(edit_sta_col)
        vbox_sta_edit.setContentsMargins(0, 0, 0, 0)
        action_sta = QWidget()
        hbox_action_sta = QHBoxLayout(action_sta)
        hbox_action_sta.setContentsMargins(0, 0, 0, 0)
        vbox_sta_edit.addWidget(action_sta)
        self.btn_add_sta = btn_add_sta = QPushButton("Add")
        btn_add_sta.setFixedSize(96, 24)
        btn_add_sta.setDisabled(True)
        hbox_action_sta.addWidget(btn_add_sta)
        hbox_action_sta.addStretch()
        self.list_obs = QListWidget()
        self.list_obs.setMinimumHeight(120)
        self.list_obs.setViewportMargins(2, 2, 17, 2)
        vbox_sta_edit.addWidget(self.list_obs)
        # vbox_sta_edit.addStretch()
        hbox_sta.addWidget(edit_sta_col)
        cbox_more.setContentLayout(vbox_more)

    def config_logic(self):
        self.dict_con = {}

        self.edit_var.textChanged.connect(self.filter_var)
        self.check_only_var.stateChanged.connect(self.toggle_var_show_mode)
        self.edit_obj.textChanged.connect(self.filter_obj)
        self.check_only_obj.stateChanged.connect(self.toggle_obj_show_mode)
        self.btn_params.toggled.connect(self.toggle_params)
        self.animation.finished.connect(self.animation_finished)

    def toggle_params(self, checked):
        if not checked:
            self.animation.setStartValue(self.edit.sizeHint().height())
            self.animation.setEndValue(0)
        else:
            # Animate to fill the available vertical space
            self.animation.setStartValue(0)
            self.animation.setEndValue(self.edit.sizeHint().height() * 4)
            self.edit.show()

        # Configure the animation
        self.animation.start()

    def animation_finished(self):
        if self.edit.maximumHeight() == 0:
            self.edit.hide()

    def toggle_var_show_mode(self, _):
        self.var_table.toggle_show_mode(self.check_only_var.isChecked())

    def add_var(self, name, lb, ub):
        self.var_table.add_variable(name, lb, ub)
        self.filter_var()

    def filter_var(self):
        keyword = self.edit_var.text()
        rx = QRegExp(keyword)

        _variables = []
        for var in self.var_table.all_variables:
            vname = next(iter(var))
            if rx.indexIn(vname, 0) != -1:
                _variables.append(var)

        self.var_table.update_variables(_variables, 1)

    def toggle_obj_show_mode(self, _):
        self.obj_table.toggle_show_mode(self.check_only_obj.isChecked())

    def filter_obj(self):
        keyword = self.edit_obj.text()
        rx = QRegExp(keyword)

        _objectives = []
        for obj in self.obj_table.all_objectives:
            oname = next(iter(obj))
            if rx.indexIn(oname, 0) != -1:
                _objectives.append(obj)

        self.obj_table.update_objectives(_objectives, 1)

    def _fit_content(self, list):
        height = list.sizeHintForRow(0) * list.count() + 2 * list.frameWidth() + 4
        height = max(28, min(height, 192))
        list.setFixedHeight(height)

    def fit_content(self):
        return

    def switch_var_panel_style(self, auto=True):
        if auto:
            self.var_panel.setStyleSheet(stylesheet_auto)
            self.msg_auto.setStyleSheet(stylesheet_auto_msg)
            self.msg_auto.setText(self.MSG_AUTO)
        else:
            self.var_panel.setStyleSheet(stylesheet_manual)
            self.msg_auto.setStyleSheet(stylesheet_manual_msg)
            self.msg_auto.setText(self.MSG_MANUAL)

    def update_stylesheets(self, environment=""):
        if environment in self.env_dict:
            color_dict = self.env_dict[environment]
            stylesheet = f"""
                #EnvBox {{
                    border-radius: 4px;
                    border-color: {color_dict['normal']};
                    border-width: 4px;
                }}
            """
        else:
            stylesheet = ""
        self.setStyleSheet(stylesheet)



from PyQt5 import QtCore, QtGui, QtWidgets


stylesheet_toolbutton = """
QToolButton
{
    border: none;
}
"""


# https://stackoverflow.com/a/56293688
class ScrollArea(QtWidgets.QScrollArea):
    resized = QtCore.pyqtSignal()

    def resizeEvent(self, e):
        self.resized.emit()
        return super(ScrollArea, self).resizeEvent(e)


# https://stackoverflow.com/a/52617714/4263605
class CollapsibleBox(QtWidgets.QWidget):
    def __init__(self, parent=None, title="", duration=100, tooltip=None):
        super(CollapsibleBox, self).__init__(parent)

        self.title = title
        self.duration = duration

        cool_font = QtGui.QFont()
        cool_font.setWeight(QtGui.QFont.DemiBold)
        cool_font.setPixelSize(13)

        self.toggle_button = QtWidgets.QToolButton(
            text=title, checkable=True, checked=False
        )
        self.toggle_button.setFont(cool_font)
        self.toggle_button.setFixedHeight(28)
        self.toggle_button.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed
        )
        self.toggle_button.setStyleSheet(stylesheet_toolbutton)
        self.toggle_button.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
        self.toggle_button.setIconSize(QtCore.QSize(11, 11))
        self.toggle_button.setArrowType(QtCore.Qt.RightArrow)
        self.toggle_button.setToolTip(tooltip)
        self.toggle_button.clicked.connect(self.start_animation)
        # self.toggle_button.setText(f'+ {title}')

        self.toggle_animation = QtCore.QParallelAnimationGroup(self)

        self.content_area = ScrollArea(maximumHeight=0, minimumHeight=0)
        self.content_area.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed
        )
        self.content_area.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.content_area.resized.connect(self.updateContentLayout)

        lay = QtWidgets.QVBoxLayout(self)
        lay.setSpacing(0)
        lay.setContentsMargins(0, 0, 0, 0)
        lay.addWidget(self.toggle_button)
        lay.addWidget(self.content_area)

        self.toggle_animation.addAnimation(
            QtCore.QPropertyAnimation(self, b"minimumHeight")
        )
        self.toggle_animation.addAnimation(
            QtCore.QPropertyAnimation(self, b"maximumHeight")
        )
        self.toggle_animation.addAnimation(
            QtCore.QPropertyAnimation(self.content_area, b"maximumHeight")
        )

    @QtCore.pyqtSlot()
    def start_animation(self):
        checked = self.toggle_button.isChecked()
        arrow_type = QtCore.Qt.DownArrow if checked else QtCore.Qt.RightArrow
        self.toggle_button.setArrowType(arrow_type)
        # self.toggle_button.setText(f'- {self.title}' if not checked else f'+ {self.title}')
        direction = (
            QtCore.QAbstractAnimation.Forward
            if checked
            else QtCore.QAbstractAnimation.Backward
        )
        self.toggle_animation.setDirection(direction)
        self.toggle_animation.start()

    def expand(self):
        if not self.toggle_button.isChecked():
            self.toggle_button.click()

    def updateContentLayout(self):
        if (
            self.toggle_button.isChecked()
            and self.toggle_animation.state() != self.toggle_animation.Running
        ):
            content_height = self.content_area.layout().sizeHint().height()
            self.setMinimumHeight(self.collapsed_height + content_height)
            self.setMaximumHeight(self.collapsed_height + content_height)
            self.content_area.setMaximumHeight(content_height)
        self.updateGeometry()
        p = self.parent()
        if isinstance(p, ScrollArea):
            p.resized.emit()

    def setContentLayout(self, layout):
        lay = self.content_area.layout()
        del lay
        self.content_area.setLayout(layout)
        self.collapsed_height = collapsed_height = (
            self.sizeHint().height() - self.content_area.maximumHeight()
        )
        content_height = layout.sizeHint().height()
        for i in range(self.toggle_animation.animationCount() - 1):
            animation = self.toggle_animation.animationAt(i)
            animation.setDuration(self.duration)
            animation.setStartValue(collapsed_height)
            animation.setEndValue(collapsed_height + content_height)

        content_animation = self.toggle_animation.animationAt(
            self.toggle_animation.animationCount() - 1
        )
        content_animation.setDuration(self.duration)
        content_animation.setStartValue(0)
        content_animation.setEndValue(content_height)



# PyQt Functionality Snippet by Apocalyptech
# "Licensed" in the Public Domain under CC0 1.0 Universal (CC0 1.0)
# Public Domain Dedication.  Use it however you like!
#
# https://creativecommons.org/publicdomain/zero/1.0/
# https://creativecommons.org/publicdomain/zero/1.0/legalcode

from PyQt5 import QtWidgets, QtGui


class MyModel(QtGui.QStandardItemModel):
    def dropMimeData(self, data, action, row, col, parent):
        """
        Always move the entire row, and don't allow column "shifting"
        """
        return super().dropMimeData(data, action, row, 0, parent)


class MyStyle(QtWidgets.QProxyStyle):
    def drawPrimitive(self, element, option, painter, widget=None):
        """
        Draw a line across the entire row rather than just the column
        we're hovering over.  This may not always work depending on global
        style - for instance I think it won't work on OSX.
        """
        if element == self.PE_IndicatorItemViewItemDrop and not option.rect.isNull():
            option_new = QtWidgets.QStyleOption(option)
            option_new.rect.setLeft(0)
            if widget:
                option_new.rect.setRight(widget.width())
            option = option_new
        super().drawPrimitive(element, option, painter, widget)


class MyTableView(QtWidgets.QTableView):
    def __init__(self, parent):
        super().__init__(parent)
        self.verticalHeader().hide()
        self.horizontalHeader().hide()
        self.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        self.setSelectionBehavior(self.SelectRows)
        self.setSelectionMode(self.SingleSelection)
        self.setShowGrid(False)
        self.setDragDropMode(self.InternalMove)
        self.setDragDropOverwriteMode(False)

        # Set our custom style - this draws the drop indicator across the whole row
        self.setStyle(MyStyle())

        # Set our custom model - this prevents row "shifting"
        self.model = MyModel()
        self.setModel(self.model)

        for idx, data in enumerate(["foo", "bar", "baz"]):
            item_1 = QtGui.QStandardItem("Item {}".format(idx))
            item_1.setEditable(False)
            item_1.setDropEnabled(False)

            item_2 = QtGui.QStandardItem(data)
            item_2.setEditable(False)
            item_2.setDropEnabled(False)

            self.model.appendRow([item_1, item_2])



import warnings
import traceback
import copy
from functools import partial
import os
import yaml

import numpy as np
import pandas as pd
from PyQt5.QtCore import Qt, pyqtSignal, QTimer
from PyQt5.QtWidgets import QLineEdit, QLabel, QPushButton, QFileDialog
from PyQt5.QtWidgets import QListWidgetItem, QMessageBox, QWidget, QTabWidget
from PyQt5.QtWidgets import QVBoxLayout, QHBoxLayout, QScrollArea
from PyQt5.QtWidgets import QTableWidgetItem, QPlainTextEdit
from coolname import generate_slug
from xopt import VOCS
from xopt.generators import (
    get_generator_defaults,
    all_generator_names,
    get_generator_dynamic,
)
from xopt.utils import get_local_region
from pydantic import ValidationError

from badger.gui.acr.components.generator_cbox import BadgerAlgoBox
from badger.gui.default.components.data_table import (
    get_table_content_as_dict,
    set_init_data_table,
    update_init_data_table,
)
from badger.gui.acr.components.env_cbox import BadgerEnvBox
from badger.gui.default.components.filter_cbox import BadgerFilterBox
from badger.gui.default.components.state_item import state_item
from badger.gui.default.windows.docs_window import BadgerDocsWindow
from badger.gui.default.windows.env_docs_window import BadgerEnvDocsWindow
from badger.gui.default.windows.edit_script_dialog import BadgerEditScriptDialog
from badger.gui.default.windows.lim_vrange_dialog import BadgerLimitVariableRangeDialog
from badger.gui.acr.windows.ind_lim_vrange_dialog import (
    BadgerIndividualLimitVariableRangeDialog,
)
from badger.gui.default.windows.review_dialog import BadgerReviewDialog
from badger.gui.default.windows.add_random_dialog import BadgerAddRandomDialog
from badger.gui.default.windows.message_dialog import BadgerScrollableMessageBox
from badger.gui.default.utils import filter_generator_config
from badger.gui.acr.components.archive_search import ArchiveSearchWidget
from badger.archive import update_run
from badger.environment import instantiate_env
from badger.errors import (
    BadgerRoutineError,
    BadgerEnvVarError,
    BadgerEnvInstantiationError,
    VariableRangeError,
)
from badger.factory import list_generators, list_env, get_env
from badger.routine import Routine
from badger.settings import init_settings
from datetime import datetime
from badger.utils import (
    get_yaml_string,
    load_config,
    strtobool,
    get_badger_version,
    get_xopt_version,
)

LABEL_WIDTH = 96


def format_validation_error(e: ValidationError) -> str:
    """Convert Pydantic ValidationError into a friendly message."""
    messages = ["\n"]
    for err in e.errors():
        loc = " -> ".join(str(item) for item in err["loc"])
        msg = f"{loc}: {err['msg']}\n"
        messages.append(msg)
    return "\n".join(messages)


class BadgerRoutinePage(QWidget):
    sig_updated = pyqtSignal(str, str)  # routine name, routine description
    sig_load_template = pyqtSignal(str)  # template path
    sig_save_template = pyqtSignal(str)  # template path

    def __init__(self):
        super().__init__()

        self.generators = list_generators()
        self.envs = list_env()
        self.env = None
        self.routine = None
        self.script = ""
        self.window_docs = BadgerDocsWindow(self, "")
        self.window_env_docs = BadgerEnvDocsWindow(self, "")
        self.vars_env = None  # needed for passing env vars to the var table

        # Limit variable ranges
        self.limit_option = {
            "limit_option_idx": 0,
            "ratio_curr": 0.1,
            "ratio_full": 0.1,
            "delta": 0.1,
        }

        # Add radom points config
        self.add_rand_config = {
            "method": 0,
            "n_points": 3,
            "fraction": 0.1,
        }
        self.rc_dialog = None

        # Record the initial table actions
        self.init_table_actions = []
        # Record the ratio var ranges
        self.ratio_var_ranges = {}
        # Record the overrided variable ranges
        self.var_hard_limit = {}

        self.init_ui()
        self.config_logic()

        # Trigger the re-rendering of the environment box
        self.env_box.relative_to_curr.setChecked(True)
        # remember user selection from lim_vrange_dialog gui
        # 2: not initialized, 1: apply to all, 0: apply to only visible
        self.lim_apply_to_vars = 2

    def init_ui(self):
        config_singleton = init_settings()

        # Set up the layout
        vbox = QVBoxLayout(self)
        vbox.setContentsMargins(8, 18, 8, 0)

        self.tabs = tabs = QTabWidget()
        vbox.addWidget(tabs)

        # Meta group
        self.group_meta = group_meta = QWidget()
        vbox_meta = QVBoxLayout(group_meta)
        vbox_meta.setContentsMargins(8, 8, 8, 8)
        tabs.addTab(group_meta, "Metadata")

        # Name
        name = QWidget()
        hbox_name = QHBoxLayout(name)
        hbox_name.setContentsMargins(0, 0, 0, 0)
        label = QLabel("Name")
        label.setFixedWidth(LABEL_WIDTH)
        self.edit_save = edit_save = QLineEdit()
        edit_save.setPlaceholderText(generate_slug(2))
        hbox_name.addWidget(label)
        hbox_name.addWidget(edit_save, 1)
        vbox_meta.addWidget(name, alignment=Qt.AlignTop)

        # Description
        descr = QWidget()
        hbox_descr = QHBoxLayout(descr)
        hbox_descr.setContentsMargins(0, 0, 0, 0)
        lbl_descr_col = QWidget()
        vbox_lbl_descr = QVBoxLayout(lbl_descr_col)
        vbox_lbl_descr.setContentsMargins(0, 0, 0, 0)
        lbl_descr = QLabel("Description")
        lbl_descr.setFixedWidth(LABEL_WIDTH)
        vbox_lbl_descr.addWidget(lbl_descr)
        vbox_lbl_descr.addStretch(1)
        hbox_descr.addWidget(lbl_descr_col)

        edit_descr_col = QWidget()
        vbox_descr_edit = QVBoxLayout(edit_descr_col)
        vbox_descr_edit.setContentsMargins(0, 0, 0, 0)
        self.edit_descr = edit_descr = QPlainTextEdit()
        edit_descr.setMinimumHeight(80)
        vbox_descr_edit.addWidget(edit_descr)
        descr_bar = QWidget()
        hbox_descr_bar = QHBoxLayout(descr_bar)
        hbox_descr_bar.setContentsMargins(0, 0, 0, 0)
        self.btn_descr_update = btn_update = QPushButton("Update Description")
        btn_update.setDisabled(True)
        btn_update.setFixedSize(128, 24)
        hbox_descr_bar.addStretch(1)
        hbox_descr_bar.addWidget(btn_update)
        vbox_descr_edit.addWidget(descr_bar)
        hbox_descr.addWidget(edit_descr_col)
        vbox_meta.addWidget(descr)
        descr_bar.hide()

        # Save Template Button
        template_button = QWidget()
        hbox_name = QHBoxLayout(template_button)
        hbox_name.setContentsMargins(0, 0, 0, 0)
        self.save_template_button = save_template_button = QPushButton(
            "Save as Template"
        )
        save_template_button.setFixedSize(128, 24)
        hbox_name.addWidget(save_template_button, alignment=Qt.AlignRight)
        vbox_meta.addWidget(template_button, alignment=Qt.AlignBottom)
        template_button.show()

        # Tags
        self.cbox_tags = cbox_tags = BadgerFilterBox(title=" Tags")
        if not strtobool(config_singleton.read_value("BADGER_ENABLE_ADVANCED")):
            cbox_tags.hide()
        vbox_meta.addWidget(cbox_tags, alignment=Qt.AlignTop)
        # vbox_meta.addStretch()

        # vbox.addWidget(group_meta)

        # Env box
        self.BADGER_PLUGIN_ROOT = BADGER_PLUGIN_ROOT = config_singleton.read_value(
            "BADGER_PLUGIN_ROOT"
        )
        env_dict_dir = os.path.join(
            BADGER_PLUGIN_ROOT, "environments", "env_colors.yaml"
        )
        try:
            with open(env_dict_dir, "r") as stream:
                env_dict = yaml.safe_load(stream)
        except (FileNotFoundError, yaml.YAMLError):
            env_dict = {}
        self.env_box = BadgerEnvBox(env_dict, None, self.envs)
        scroll_area = QScrollArea()
        scroll_area.setFrameShape(QScrollArea.NoFrame)
        scroll_area.setStyleSheet(
            """
            QScrollArea {
                border: none;  /* Remove border */
                margin: 0px;   /* Remove margin */
                padding: 0px;  /* Remove padding */
            }
            QScrollArea > QWidget {
                margin: 0px;   /* Remove margin inside */
            }
        """
        )
        scroll_content_env = QWidget()
        scroll_layout_env = QVBoxLayout(scroll_content_env)
        scroll_layout_env.setContentsMargins(0, 0, 15, 0)
        scroll_layout_env.addWidget(self.env_box)
        scroll_area.setWidgetResizable(True)
        scroll_area.setWidget(scroll_content_env)
        tabs.addTab(scroll_area, "Environment + VOCS")

        # Algo box
        self.generator_box = BadgerAlgoBox(None, self.generators)
        tabs.addTab(self.generator_box, "Algorithm")

        tabs.setCurrentIndex(1)  # Show the env box by default

        # vbox.addStretch()

        # Template path
        try:
            self.template_dir = config_singleton.read_value("BADGER_TEMPLATE_ROOT")
        except KeyError:
            self.template_dir = os.path.join(self.BADGER_PLUGIN_ROOT, "templates")

    def config_logic(self):
        self.btn_descr_update.clicked.connect(self.update_description)
        self.env_box.load_template_button.clicked.connect(self.load_template_yaml)
        self.save_template_button.clicked.connect(self.save_template_yaml)
        self.generator_box.cb.currentIndexChanged.connect(self.select_generator)
        self.generator_box.btn_docs.clicked.connect(self.open_generator_docs)
        self.generator_box.check_use_script.stateChanged.connect(self.toggle_use_script)
        self.generator_box.btn_edit_script.clicked.connect(self.edit_script)
        self.env_box.cb.currentIndexChanged.connect(self.select_env)
        self.env_box.btn_env_play.clicked.connect(self.open_playground)
        self.env_box.btn_pv.clicked.connect(self.open_archive_search)
        self.env_box.btn_docs.clicked.connect(self.open_environment_docs)
        self.env_box.btn_add_var.clicked.connect(self.add_var)
        self.env_box.btn_lim_vrange.clicked.connect(self.limit_variable_ranges)
        self.env_box.btn_add_con.clicked.connect(self.add_constraint)
        self.env_box.btn_add_sta.clicked.connect(self.add_state)
        self.env_box.btn_add_curr.clicked.connect(
            partial(self.fill_curr_in_init_table, record=True)
        )
        self.env_box.btn_add_rand.clicked.connect(self.show_add_rand_dialog)
        self.env_box.btn_clear.clicked.connect(
            partial(self.clear_init_table, reset_actions=True)
        )
        self.env_box.btn_add_row.clicked.connect(self.add_row_to_init_table)
        self.env_box.relative_to_curr.stateChanged.connect(self.toggle_relative_to_curr)
        self.env_box.btn_refresh.clicked.connect(self.refresh_variables)
        self.env_box.var_table.sig_sel_changed.connect(self.update_init_table)
        self.env_box.var_table.sig_pv_added.connect(self.handle_pv_added)
        self.env_box.var_table.sig_var_config.connect(self.handle_var_config)

    def load_template_yaml(self) -> None:
        """
        Load data from template .yaml into template_dict dictionary.
        This function expects to be called via an action from
        a QPushButton
        """

        if isinstance(self.sender(), QPushButton):
            # load template from button
            options = QFileDialog.Options()
            template_path, _ = QFileDialog.getOpenFileName(
                self,
                "Select YAML File",
                self.template_dir,
                "YAML Files (*.yaml);;All Files (*)",
                options=options,
            )

        if not template_path:
            return

        # Load template file
        try:
            with open(template_path, "r") as stream:
                template_dict = yaml.safe_load(stream)
                self.set_options_from_template(template_dict=template_dict)
                self.sig_load_template.emit(
                    f"Options loaded from template: {os.path.basename(template_path)}"
                )
        except (FileNotFoundError, yaml.YAMLError) as e:
            print(f"Error loading template: {e}")
            return

    def set_options_from_template(self, template_dict: dict):
        """
        Fills in routine_page GUI with relevant info from template_dict
        dictionary
        """

        # Compose the template
        try:
            name = template_dict["name"]
            description = template_dict["description"]
            relative_to_current = template_dict["relative_to_current"]
            generator_name = template_dict["generator"]["name"]
            env_name = template_dict["environment"]["name"]
            vrange_limit_options = template_dict["vrange_limit_options"]
            try:  # this one is optional
                vrange_hard_limit = template_dict["vrange_hard_limit"]
            except KeyError:
                vrange_hard_limit = {}
            initial_point_actions = template_dict[
                "initial_point_actions"
            ]  # should be type: add_curr
            critical_constraint_names = template_dict["critical_constraint_names"]
            env_params = template_dict["environment"]["params"]
        except KeyError as e:
            QMessageBox.warning(self, "Error", f"Missing key in template: {e}")
            return

        # set vocs
        vocs = VOCS(
            variables=template_dict["vocs"]["variables"],
            objectives=template_dict["vocs"]["objectives"],
            constraints=template_dict["vocs"]["constraints"],
            constants={},
            observables=template_dict["vocs"]["observables"],
        )

        # set name
        self.edit_save.setText(name)

        # set description
        self.edit_descr.setPlainText(description)

        # set generator
        if generator_name in self.generators:
            i = self.generators.index(generator_name)
            self.generator_box.cb.setCurrentIndex(i)

            filtered_config = filter_generator_config(
                generator_name, template_dict["generator"]
            )
            self.generator_box.edit.setPlainText(get_yaml_string(filtered_config))

        # set environment
        if env_name in self.envs:
            i = self.envs.index(env_name)
            self.env_box.cb.setCurrentIndex(i)
            self.env_box.edit.setPlainText(get_yaml_string(env_params))

        # Load the vrange options and hard limits
        self.ratio_var_ranges = vrange_limit_options
        self.init_table_actions = initial_point_actions
        self.var_hard_limit = vrange_hard_limit

        self.env_box.check_only_var.setChecked(True)

        # Add additional variables to table as well
        # Combine the variables from the env with the additional variables
        all_variables = {}  # note this stores the hard bounds of the variables
        try:
            additional_variables = template_dict["additional_variables"]
        except KeyError:
            additional_variables = []  # init to empty list if not present
        if self.vars_env:
            for i in self.vars_env:
                all_variables.update(i)
        if additional_variables:  # there are additional variables
            env = self.create_env()
            for vname in additional_variables:
                try:
                    bounds = env._get_bounds([vname])[vname]
                except BadgerEnvVarError as e:
                    msg = str(e)
                    bounds = eval(msg.split(": ")[1])

                all_variables.update({vname: bounds})
        # Override the hard limits with the ones from the routine
        all_variables.update(self.var_hard_limit)
        # Format for update_variables method
        all_variables = dict(sorted(all_variables.items()))
        all_variables = [{key: value} for key, value in all_variables.items()]

        self.env_box.var_table.update_variables(all_variables)
        self.env_box.var_table.set_selected(vocs.variables)
        self.env_box.var_table.addtl_vars = additional_variables

        flag_relative = relative_to_current
        self.env_box.relative_to_curr.blockSignals(True)
        self.env_box.relative_to_curr.setChecked(flag_relative)
        self.env_box.relative_to_curr.blockSignals(False)
        self.toggle_relative_to_curr(flag_relative, refresh=False)

        if env_name:
            if flag_relative:
                bounds = self.calc_auto_bounds()
                self.env_box.var_table.set_bounds(bounds, signal=False)
            else:
                self.env_box.var_table.set_bounds(vocs.variables, signal=False)
            # Populate the initial table anyways, auto mode or not
            self.clear_init_table(reset_actions=False)
            self.update_init_table(force=True)

        # set objectives
        try:
            formulas = template_dict["formulas"]
        except KeyError:
            formulas = {}

        # Initialize the objective table with env observables
        objectives = []
        for name in self.configs["observations"]:
            obj = {name: "MINIMIZE"}
            objectives.append(obj)
        self.env_box.obj_table.update_objectives(objectives)

        for name, formula in formulas.items():
            formula_tuple = (
                name,
                formula["formula"],
                formula["variable_mapping"],
            )
            self.env_box.obj_table.add_formula_objective(formula_tuple)
        self.env_box.obj_table.set_selected(vocs.objectives)
        self.env_box.obj_table.set_rules(vocs.objectives)
        self.env_box.check_only_obj.setChecked(True)

        # set constraints
        self.env_box.con_table.clear_constraints()
        constraints = vocs.constraints
        if len(constraints):
            for name, val in constraints.items():
                relation, thres = val
                critical = name in critical_constraint_names
                relation = ["GREATER_THAN", "LESS_THAN"].index(relation)
                self.add_constraint(name, relation, thres, critical)

        # set observables
        self.env_box.list_obs.clear()
        observables = vocs.observable_names
        if len(observables):
            for name_sta in observables:
                self.add_state(name_sta)

    def generate_template_dict_from_gui(self):
        """
        Generate a template dictionary from the current state of the GUI
        """

        vocs, critical_constraints = self._compose_vocs()

        # Filter generator
        generator_name = self.generator_box.cb.currentText()

        generator_config = self._filter_generator_params(
            generator_name=generator_name,
            generator_config=load_config(self.generator_box.edit.toPlainText()),
        )

        template_dict = {
            "name": self.edit_save.text(),
            "description": str(self.edit_descr.toPlainText()),
            "relative_to_current": self.env_box.relative_to_curr.isChecked(),
            "generator": {
                "name": generator_name,
            }
            | generator_config,
            "environment": {
                "name": self.env_box.cb.currentText(),
                "params": load_config(self.env_box.edit.toPlainText()),
            },
            "vrange_limit_options": self.ratio_var_ranges,
            "vrange_hard_limit": self.var_hard_limit,
            "additional_variables": self.env_box.var_table.addtl_vars,
            "formulas": self.env_box.obj_table.formulas,
            "initial_point_actions": self.init_table_actions,
            "critical_constraint_names": critical_constraints,
            "vocs": vars(vocs),
            "badger_version": get_badger_version(),
            "xopt_version": get_xopt_version(),
        }

        return template_dict

    def _filter_generator_params(self, generator_name: str, generator_config: dict):
        """
        Filter which generator parameters get saved to template
        """

        if generator_name in ["expected_improvement", "upper_confidence_bound"]:
            if (
                "turbo_controller" in generator_config
                and generator_config["turbo_controller"] is not None
                and isinstance(generator_config["turbo_controller"], dict)
            ):
                turbo = generator_config["turbo_controller"]
                generator_config["turbo_controller"] = {
                    k: v
                    for k, v in turbo.items()
                    if k
                    in {
                        "name",
                        "length",
                        "length_max",
                        "length_min",
                        "failure_tolerance",
                        "success_tolerance",
                        "scale_factor",
                        "restrict_model_data",
                    }
                }

        return generator_config

    def save_template_yaml(self):
        """
        Save the current routine as a template .yaml file
        """

        template_dict = self.generate_template_dict_from_gui()

        options = QFileDialog.Options()
        # Suggest a filename based on the routine name or placeholder
        routine_name = self.edit_save.text() or self.edit_save.placeholderText()
        if not routine_name:
            routine_name = "template_" + datetime.now().strftime("%y%m%d_%H%M%S")
        suggested_filename = f"{routine_name}.yaml"
        template_path, _ = QFileDialog.getSaveFileName(
            self,
            "Save Template",
            os.path.join(self.template_dir, suggested_filename),
            "YAML Files (*.yaml);;All Files (*)",
            options=options,
        )

        if not template_path:
            return

        try:
            with open(template_path, "w") as stream:
                yaml.dump(template_dict, stream)
                self.sig_save_template.emit(
                    f"Current routine options saved to template: {os.path.basename(template_path)}"
                )
        except (FileNotFoundError, yaml.YAMLError) as e:
            print(f"Error saving template: {e}")
            return

    def refresh_ui(self, routine: Routine = None, silent: bool = False):
        self.routine = routine  # save routine for future reference

        self.generators = list_generators()
        self.envs = list_env()
        # Clean up the constraints/observables list
        self.env_box.con_table.clear_constraints()
        self.env_box.list_obs.clear()

        if routine is None:
            # Reset the generator and env configs
            self.generator_box.cb.setCurrentIndex(-1)
            self.env_box.cb.setCurrentIndex(-1)
            init_table = self.env_box.init_table
            init_table.clear()
            init_table.horizontalHeader().setVisible(False)
            init_table.setRowCount(10)
            init_table.setColumnCount(0)

            # Reset the routine configs check box status
            self.env_box.check_only_var.setChecked(False)
            self.env_box.check_only_obj.setChecked(False)
            self.env_box.relative_to_curr.setChecked(True)
            self.try_populate_init_table()

            # Reset the save settings
            name = generate_slug(2)
            self.edit_save.setText("")
            self.edit_save.setPlaceholderText(name)
            self.edit_descr.setPlainText("")
            self.btn_descr_update.setDisabled(True)

            return

        # Enable description edition
        self.btn_descr_update.setDisabled(False)
        # Fill in the generator and env configs
        name_generator = routine.generator.name
        try:
            idx_generator = self.generators.index(name_generator)
        except ValueError as e:
            if not silent:  # show the error message if not in silent mode
                details = traceback.format_exc()
                dialog = BadgerScrollableMessageBox(
                    title="Error!", text=str(e), parent=self
                )
                dialog.setIcon(QMessageBox.Critical)
                dialog.setDetailedText(details)
                dialog.exec_()

            idx_generator = -1

        self.generator_box.cb.setCurrentIndex(idx_generator)
        # self.generator_box.edit.setPlainText(routine.generator.yaml())
        filtered_config = filter_generator_config(
            name_generator, routine.generator.model_dump()
        )
        self.generator_box.edit.setPlainText(get_yaml_string(filtered_config))
        self.script = routine.script

        name_env = routine.environment.name
        idx_env = self.envs.index(name_env)
        self.env_box.cb.setCurrentIndex(idx_env)
        env_params = routine.environment.model_dump()
        del env_params["interface"]
        self.env_box.edit.setPlainText(get_yaml_string(env_params))

        # Config the vocs panel
        variables = routine.vocs.variable_names
        self.env_box.check_only_var.setChecked(True)

        self.env_box.edit_var.clear()

        try:
            self.var_hard_limit = routine.vrange_hard_limit
        except AttributeError:
            self.var_hard_limit = {}
        # Add additional variables to table as well
        # Combine the variables from the env with the additional variables
        all_variables = {}  # note this stores the hard bounds of the variables
        for i in self.vars_env:
            all_variables.update(i)
        if routine.additional_variables:  # there are additional variables
            env = self.create_env()
            for vname in routine.additional_variables:
                try:
                    bounds = env._get_bounds([vname])[vname]
                except BadgerEnvVarError as e:
                    msg = str(e)
                    bounds = eval(msg.split(": ")[1])

                all_variables.update({vname: bounds})
        # Override the hard limits with the ones from the routine
        all_variables.update(self.var_hard_limit)
        # Format for update_variables method
        all_variables = dict(sorted(all_variables.items()))
        all_variables = [{key: value} for key, value in all_variables.items()]

        self.env_box.var_table.update_variables(all_variables)
        self.env_box.var_table.set_selected(variables)
        self.env_box.var_table.addtl_vars = routine.additional_variables

        flag_relative = routine.relative_to_current
        if flag_relative:  # load the relative to current settings
            self.ratio_var_ranges = routine.vrange_limit_options
            self.init_table_actions = routine.initial_point_actions
        self.env_box.relative_to_curr.blockSignals(True)
        self.env_box.relative_to_curr.setChecked(flag_relative)
        self.env_box.relative_to_curr.blockSignals(False)
        self.toggle_relative_to_curr(flag_relative, refresh=False)

        # Always use ranges stored in routine
        self.env_box.var_table.set_bounds(routine.vocs.variables, signal=False)

        # Fill in initial points stored in routine if available
        try:
            # Update the header
            update_init_data_table(self.env_box.init_table, variables)
            # Fill in the table
            init_points = routine.initial_points
            set_init_data_table(self.env_box.init_table, init_points)
        except KeyError:
            set_init_data_table(self.env_box.init_table, None)

        # set objectives
        try:
            formulas = routine.formulas
        except AttributeError:
            formulas = {}

        # Initialize the objective table with env observables
        objectives = []
        for name in self.configs["observations"]:
            obj = {name: "MINIMIZE"}
            objectives.append(obj)
        self.env_box.obj_table.update_objectives(objectives)

        for name, formula in formulas.items():
            formula_tuple = (
                name,
                formula["formula"],
                formula["variable_mapping"],
            )
            self.env_box.obj_table.add_formula_objective(formula_tuple)
        self.env_box.obj_table.set_selected(routine.vocs.objectives)
        self.env_box.obj_table.set_rules(routine.vocs.objectives)
        self.env_box.check_only_obj.setChecked(True)

        constraints = routine.vocs.constraints
        if len(constraints):
            for name, val in constraints.items():
                relation, thres = val
                critical = name in routine.critical_constraint_names
                relation = ["GREATER_THAN", "LESS_THAN", "EQUAL_TO"].index(relation)
                self.add_constraint(name, relation, thres, critical)

        observables = routine.vocs.observable_names
        if len(observables):
            for name_sta in observables:
                self.add_state(name_sta)

        # Config the metadata
        self.edit_save.setPlaceholderText(generate_slug(2))
        self.edit_save.setText(routine.name)
        self.edit_descr.setPlainText(routine.description)

        self.generator_box.check_use_script.setChecked(not not self.script)

    def select_generator(self, i):
        # Reset the script
        self.script = ""
        self.generator_box.check_use_script.setChecked(False)

        if i == -1:
            self.generator_box.edit.setPlainText("")
            self.generator_box.cb_scaling.setCurrentIndex(-1)
            return

        name = self.generators[i]
        default_config = get_generator_defaults(name)

        if name in all_generator_names["bo"]:
            # Patch for BOs that make the low noise prior False by default
            default_config["gp_constructor"]["use_low_noise_prior"] = False
            # Patch for BOs that turn on TuRBO by default
            # default_config["turbo_controller"] = "optimize"

        # Patch to only show part of the config
        filtered_config = filter_generator_config(name, default_config)
        self.generator_box.edit.setPlainText(get_yaml_string(filtered_config))

        # Update the docs
        self.window_docs.update_docs(name)

    def _fill_init_table(self):  # make sure self.init_table_actions is set
        for action in self.init_table_actions:
            if action["type"] == "add_curr":
                self.fill_curr_in_init_table(record=False)
            elif action["type"] == "add_rand":
                try:
                    self.add_rand_in_init_table(
                        add_rand_config=action["config"],
                        record=False,
                    )
                except IndexError:  # lower bound is the same as upper bound
                    pass

    def toggle_use_script(self):
        if self.generator_box.check_use_script.isChecked():
            self.generator_box.btn_edit_script.show()
            self.generator_box.edit.setReadOnly(True)
            self.refresh_params_generator()
        else:
            self.generator_box.btn_edit_script.hide()
            self.generator_box.edit.setReadOnly(False)

    def edit_script(self):
        generator = self.generator_box.cb.currentText()
        dlg = BadgerEditScriptDialog(self, generator, self.script, self.script_updated)
        dlg.exec()

    def script_updated(self, text):
        self.script = text
        self.refresh_params_generator()

    def create_env(self):
        env_params = load_config(self.env_box.edit.toPlainText())
        try:
            intf_name = self.configs["interface"][0]
        except KeyError:
            intf_name = None
        configs = {"params": env_params, "interface": [intf_name]}
        try:
            env = instantiate_env(self.env, configs)
        except Exception as e:
            raise BadgerEnvInstantiationError(f"Failed to instantiate environment: {e}")

        return env

    def refresh_params_generator(self):
        if not self.script:
            return

        try:
            tmp = {}
            exec(self.script, tmp)
            try:
                tmp["generate"]  # test if generate function is defined
            except Exception as e:
                QMessageBox.warning(
                    self, "Please define a valid generate function!", str(e)
                )
                return

            env = self.create_env()
            # Get vocs
            try:
                vocs = self._compose_vocs()
            except Exception:
                vocs = None
            # Function generate comes from the script
            params_generator = tmp["generate"](env, vocs)
            self.generator_box.edit.setPlainText(get_yaml_string(params_generator))
        except Exception as e:
            QMessageBox.warning(self, "Invalid script!", str(e))

    def select_env(self, i):
        # Reset the initial table actions and ratio var ranges
        self.init_table_actions = []
        self.ratio_var_ranges = {}
        self.var_hard_limit = {}

        if hasattr(self, "archive_search"):
            self.archive_search.close()

        if i == -1:
            self.env_box.edit.setPlainText("")
            self.env_box.edit_var.clear()
            self.env_box.var_table.update_variables(None)
            self.env_box.edit_obj.clear()
            self.env_box.obj_table.update_objectives(None)
            self.configs = None
            self.env = None
            self.env_box.btn_add_con.setDisabled(True)
            self.env_box.btn_add_sta.setDisabled(True)
            self.env_box.btn_add_var.setDisabled(True)
            self.env_box.btn_lim_vrange.setDisabled(True)
            self.env_box.btn_refresh.setDisabled(True)
            self.routine = None
            self.env_box.update_stylesheets()
            return

        name = self.envs[i]
        try:
            env, configs = get_env(name)
            self.configs = configs
            self.env = env
            self.env_box.edit_var.clear()
            self.env_box.edit_obj.clear()
            self.env_box.btn_add_con.setDisabled(False)
            self.env_box.btn_add_sta.setDisabled(False)
            self.env_box.btn_add_var.setDisabled(False)
            self.env_box.btn_lim_vrange.setDisabled(False)
            self.env_box.btn_refresh.setDisabled(False)
            if self.generator_box.check_use_script.isChecked():
                self.refresh_params_generator()
        except Exception:
            self.configs = None
            self.env = None
            self.env_box.cb.setCurrentIndex(-1)
            self.env_box.btn_add_con.setDisabled(True)
            self.env_box.btn_add_sta.setDisabled(True)
            self.env_box.btn_add_var.setDisabled(True)
            self.env_box.btn_lim_vrange.setDisabled(True)
            self.routine = None
            return QMessageBox.critical(self, "Error!", traceback.format_exc())

        self.env_box.edit.setPlainText(get_yaml_string(configs["params"]))

        # Get and save vars to combine with additional vars added on the fly
        vars_env = self.vars_env = configs["variables"]
        vars_combine = [*vars_env]

        self.env_box.check_only_var.blockSignals(True)
        self.env_box.check_only_var.setChecked(False)
        self.env_box.var_table.checked_only = False  # reset the checked only flag
        self.env_box.check_only_var.blockSignals(False)
        self.env_box.var_table.update_variables(vars_combine)
        # Auto apply the limited variable ranges if the option is set
        if self.env_box.relative_to_curr.isChecked():
            self.set_vrange()

        # Needed for getting bounds on the fly
        self.env_box.var_table.env_class, self.env_box.var_table.configs = (
            self.add_var()
        )

        _objs_env = configs["observations"]
        objs_env = []
        for name in _objs_env:
            obj = {}
            obj[name] = "MINIMIZE"  # default rule
            objs_env.append(obj)
        self.env_box.check_only_obj.setChecked(False)
        self.env_box.obj_table.update_objectives(objs_env)

        self.env_box.con_table.clear_constraints()
        self.env_box.list_obs.clear()
        self.env_box.fit_content()
        # self.routine = None

        self.env_box.update_stylesheets(env.name)

        # Update the docs
        self.window_env_docs.update_docs(env.name)

    def get_init_table_header(self):
        table = self.env_box.init_table
        header_list = []
        for col in range(table.columnCount()):
            item = table.horizontalHeaderItem(col)
            if item:
                header_list.append(item.text())
            else:
                header_list.append("")  # Handle the case where the header item is None
        return header_list

    def fill_curr_in_init_table(self, record=False):
        env = self.create_env()
        table = self.env_box.init_table
        vname_selected = self.get_init_table_header()
        var_curr = env._get_variables(vname_selected)

        # Iterate through the rows
        for row in range(table.rowCount()):
            # Check if the row is empty
            if np.all(
                [not table.item(row, col).text() for col in range(table.columnCount())]
            ):
                # Fill the row with content_list
                for col, name in enumerate(vname_selected):
                    item = QTableWidgetItem(f"{var_curr[name]:.4g}")
                    table.setItem(row, col, item)
                break  # Stop after filling the first non-empty row

        if record and self.env_box.relative_to_curr.isChecked():
            self.init_table_actions.append({"type": "add_curr"})

    def save_add_rand_config(self, add_rand_config):
        self.add_rand_config = add_rand_config

    def add_rand_in_init_table(self, add_rand_config=None, record=True):
        if add_rand_config is None:
            add_rand_config = self.add_rand_config

        # Get current point
        env = self.create_env()
        vname_selected = self.get_init_table_header()
        var_curr = env._get_variables(vname_selected)

        # get small region around current point to sample
        try:
            vocs, _ = self._compose_vocs()
        except Exception:
            # Switch to manual mode to allow the user fixing the vocs issue
            QMessageBox.warning(
                self,
                "Variable range is not valid!",
                "Auto mode disabled due to invalid variable range. Please fix it before enabling auto mode.",
            )
            return self.env_box.relative_to_curr.setChecked(False)

        n_point = add_rand_config["n_points"]
        fraction = add_rand_config["fraction"]
        random_sample_region = get_local_region(var_curr, vocs, fraction=fraction)
        with warnings.catch_warnings(record=True) as caught_warnings:
            try:
                random_points = vocs.random_inputs(
                    n_point, custom_bounds=random_sample_region
                )
            except ValueError:
                raise VariableRangeError(
                    "Current value is not within variable range!\n"
                    "This is likely due to the hard variable bounds being overridden, "
                    "please examine the individual variable settings."
                )

            for warning in caught_warnings:
                # Ignore runtime warnings (usually caused by clip by bounds)
                if warning.category is RuntimeWarning:
                    pass
                else:
                    print(f"Caught user warning: {warning.message}")

        # Add points to the table
        table = self.env_box.init_table
        for row in range(table.rowCount()):
            # Check if the row is empty
            if np.all(
                [not table.item(row, col).text() for col in range(table.columnCount())]
            ):
                # Fill the row with content_list
                try:
                    point = random_points.pop(0)
                    for col, name in enumerate(vname_selected):
                        item = QTableWidgetItem(f"{point[name]:.4g}")
                        table.setItem(row, col, item)
                except IndexError:  # No more points to add
                    break

        if record and self.env_box.relative_to_curr.isChecked():
            self.init_table_actions.append(
                {
                    "type": "add_rand",
                    "config": add_rand_config,
                }
            )

    def show_add_rand_dialog(self):
        dlg = BadgerAddRandomDialog(
            self,
            self.add_rand_in_init_table,
            self.save_add_rand_config,
            self.add_rand_config,
        )
        self.rc_dialog = dlg
        try:
            dlg.exec()
        finally:
            self.rc_dialog = None

    def clear_init_table(self, reset_actions=True):
        table = self.env_box.init_table
        for row in range(table.rowCount()):
            for col in range(table.columnCount()):
                item = table.item(row, col)
                if item:
                    item.setText("")  # Set the cell content to an empty string

        if reset_actions and self.env_box.relative_to_curr.isChecked():
            self.init_table_actions = []  # reset the recorded actions

    def add_row_to_init_table(self):
        table = self.env_box.init_table
        row_position = table.rowCount()
        table.insertRow(row_position)

        for col in range(table.columnCount()):
            item = QTableWidgetItem("")
            table.setItem(row_position, col, item)

    def open_playground(self):
        pass

    def open_generator_docs(self):
        self.window_docs.show()

    def open_environment_docs(self):
        self.window_env_docs.show()

    def open_archive_search(self):
        if not hasattr(self, "archive_search") or not self.archive_search.isVisible():
            try:
                env = self.create_env()
            except AttributeError:
                raise BadgerRoutineError("No environment selected!")

            self.archive_search = ArchiveSearchWidget(environment=env)
            self.archive_search.show()
        else:
            self.archive_search.raise_()
            self.archive_search.activateWindow()

    def add_var(self):
        # TODO: Use a cached env
        env_params = load_config(self.env_box.edit.toPlainText())
        try:
            intf_name = self.configs["interface"][0]
        except KeyError:
            intf_name = None
        configs = {"params": env_params, "interface": [intf_name]}

        return self.env, configs
        # dlg = BadgerVariableDialog(self, self.env, configs, self.add_var_to_list)
        # dlg.exec()

    def limit_variable_ranges(self):
        if self.lim_apply_to_vars == 2:
            # Initialize the lim_apply_to_vars to 0 (set only visible vars)
            self.lim_apply_to_vars = 0

        dlg = BadgerLimitVariableRangeDialog(
            self,
            self.set_vrange,
            self.save_limit_option,
            self.limit_option,
            self.lim_apply_to_vars,
        )
        dlg.exec()

    def set_vrange(self, set_all=True):
        # By default update all variables no matter if selected or not
        vname_selected = []
        vrange = {}

        if set_all:
            # Set vranges for all variables
            _variables = self.env_box.var_table.all_variables
        else:
            # Only set vranges for the visible variables
            _variables = self.env_box.var_table.get_visible_variables(
                self.env_box.var_table.variables
            )

        for var in _variables:
            name = next(iter(var))
            # Set vrange no matter if selected or not
            # if set_all or self.env_box.var_table.is_checked(name):
            vname_selected.append(name)
            vrange[name] = var[name]

        env = self.create_env()
        var_curr = env._get_variables(vname_selected)

        option_idx = self.limit_option["limit_option_idx"]
        # 0: ratio with current value, 1: ratio with full range, 2: delta around current value
        if option_idx == 1:
            ratio = self.limit_option["ratio_full"]
            for i, name in enumerate(vname_selected):
                hard_bounds = vrange[name]
                delta = 0.5 * ratio * (hard_bounds[1] - hard_bounds[0])
                bounds = [var_curr[name] - delta, var_curr[name] + delta]
                bounds = np.clip(bounds, hard_bounds[0], hard_bounds[1]).tolist()
                vrange[name] = bounds
        elif option_idx == 2:
            delta = self.limit_option["delta"]
            for i, name in enumerate(vname_selected):
                hard_bounds = vrange[name]
                bounds = [var_curr[name] - delta, var_curr[name] + delta]
                bounds = np.clip(bounds, hard_bounds[0], hard_bounds[1]).tolist()
                vrange[name] = bounds
        else:
            ratio = self.limit_option["ratio_curr"]
            for i, name in enumerate(vname_selected):
                hard_bounds = vrange[name]
                sign = np.sign(var_curr[name])
                bounds = [
                    var_curr[name] * (1 - 0.5 * sign * ratio),
                    var_curr[name] * (1 + 0.5 * sign * ratio),
                ]
                bounds = np.clip(bounds, hard_bounds[0], hard_bounds[1]).tolist()
                vrange[name] = bounds

        self.env_box.var_table.set_bounds(vrange)
        self.clear_init_table(reset_actions=False)  # clear table after changing ranges
        self.update_init_table()  # auto populate if option is set

        # remember user selection for applying limit changes
        if not self.lim_apply_to_vars == 2:
            # Check if lim_apply_to_vars has been initialized
            # It will be set to 2 until the btn_lim_vrange is clicked
            self.lim_apply_to_vars = set_all

        # Record the ratio var ranges
        for vname in vname_selected:
            self.ratio_var_ranges[vname] = copy.deepcopy(self.limit_option)

    def set_ind_vrange(self, vname, config):
        hard_bounds = [config["lower_bound"], config["upper_bound"]]
        option = {
            "limit_option_idx": config["limit_option_idx"],
            "ratio_full": config["ratio_full"],
            "ratio_curr": config["ratio_curr"],
            "delta": config["delta"],
        }

        option_idx = option["limit_option_idx"]

        env = self.create_env()
        curr = env._get_variables([vname])[vname]

        # 0: ratio with current value, 1: ratio with full range, 2: delta around current value
        if option_idx == 1:
            ratio = option["ratio_full"]
            delta = 0.5 * ratio * (hard_bounds[1] - hard_bounds[0])
            bounds = [curr - delta, curr + delta]
            bounds = np.clip(bounds, hard_bounds[0], hard_bounds[1]).tolist()
        elif option_idx == 2:
            delta = option["delta"]
            bounds = [curr - delta, curr + delta]
            bounds = np.clip(bounds, hard_bounds[0], hard_bounds[1]).tolist()
        else:
            ratio = option["ratio_curr"]
            sign = np.sign(curr)
            bounds = [
                curr * (1 - 0.5 * sign * ratio),
                curr * (1 + 0.5 * sign * ratio),
            ]
            bounds = np.clip(bounds, hard_bounds[0], hard_bounds[1]).tolist()

        # Set the bounds in the table
        self.env_box.var_table.refresh_variable(vname, bounds, hard_bounds)
        self.clear_init_table(reset_actions=False)  # clear table after changing ranges
        self.update_init_table()  # auto populate if option is set

        # Record the hard bounds for the variable
        self.var_hard_limit[vname] = hard_bounds
        # Record the ratio var ranges
        self.ratio_var_ranges[vname] = copy.deepcopy(option)

    def save_limit_option(self, limit_option):
        self.limit_option = limit_option

    def add_var_to_list(self, name, lb, ub):
        # Check if already in the list
        ok = False
        try:
            self.env_box.var_table.bounds[name]
        except KeyError:
            ok = True
        if not ok:
            QMessageBox.warning(
                self, "Variable already exists!", f"Variable {name} already exists!"
            )
            return 1

        self.env_box.add_var(name, lb, ub)
        return 0

    def update_init_table(self, force=False):
        selected = self.env_box.var_table.selected
        variable_names = [v for v in selected if selected[v]]
        update_init_data_table(self.env_box.init_table, variable_names)

        if (not force) and (not self.env_box.relative_to_curr.isChecked()):
            return

        # Auto populate the initial table based on recorded actions
        if not self.init_table_actions:
            self.init_table_actions = [
                {"type": "add_curr"},
                {"type": "add_rand", "config": self.add_rand_config},
            ]
        self.clear_init_table(reset_actions=False)
        self._fill_init_table()

    def calc_auto_bounds(self):
        vname_selected = []
        vrange = {}

        for var in self.env_box.var_table.variables:
            name = next(iter(var))
            vname_selected.append(name)
            try:  # get the hard limit from the routine
                vrange[name] = self.var_hard_limit[name]
            except KeyError:
                vrange[name] = var[name]

        env = self.create_env()
        var_curr = env._get_variables(vname_selected)

        for name in vname_selected:
            try:
                limit_option = self.ratio_var_ranges[name]
            except KeyError:
                limit_option = self.limit_option

            option_idx = limit_option["limit_option_idx"]
            # 0: ratio with current value, 1: ratio with full range, 2: delta around current value
            if option_idx == 1:
                ratio = limit_option["ratio_full"]
                hard_bounds = vrange[name]
                delta = 0.5 * ratio * (hard_bounds[1] - hard_bounds[0])
                bounds = [var_curr[name] - delta, var_curr[name] + delta]
                bounds = np.clip(bounds, hard_bounds[0], hard_bounds[1]).tolist()
                vrange[name] = bounds
            elif option_idx == 2:
                delta = limit_option["delta"]
                hard_bounds = vrange[name]
                bounds = [var_curr[name] - delta, var_curr[name] + delta]
                bounds = np.clip(bounds, hard_bounds[0], hard_bounds[1]).tolist()
                vrange[name] = bounds
            else:
                ratio = limit_option["ratio_curr"]
                hard_bounds = vrange[name]
                sign = np.sign(var_curr[name])
                bounds = [
                    var_curr[name] * (1 - 0.5 * sign * ratio),
                    var_curr[name] * (1 + 0.5 * sign * ratio),
                ]
                bounds = np.clip(bounds, hard_bounds[0], hard_bounds[1]).tolist()
                vrange[name] = bounds

        return vrange

    def toggle_relative_to_curr(self, checked, refresh=True):
        if checked:
            try:
                _ = self._compose_vocs()
            except Exception:
                # Switch to manual mode to allow the user fixing the vocs issue
                # Schedule the checkbox to be clicked after the event loop finishes
                QTimer.singleShot(0, lambda: self.env_box.relative_to_curr.click())
                QMessageBox.warning(
                    self,
                    "Variable range is not valid!",
                    "Please fix the invalid variable range before enabling auto mode.",
                )
                return

            self.env_box.switch_var_panel_style(True)

            if refresh and self.env_box.var_table.selected:
                bounds = self.calc_auto_bounds()
                self.env_box.var_table.set_bounds(bounds)
                self.clear_init_table(reset_actions=False)
                # Auto populate the initial table
                self.try_populate_init_table()

            self.env_box.var_table.lock_bounds()
            self.env_box.init_table.set_uneditable()
        else:
            self.env_box.switch_var_panel_style(False)

            self.env_box.var_table.unlock_bounds()
            self.env_box.init_table.set_editable()

    def refresh_variables(self):
        variables = self.env_box.var_table.export_variables()
        bounds = self.calc_auto_bounds()

        no_need_to_update = True
        for vname in variables:
            if not np.allclose(bounds[vname], variables[vname]):
                no_need_to_update = False
                break
        if no_need_to_update:
            return

        self.env_box.var_table.set_bounds(bounds)
        self.clear_init_table(reset_actions=False)
        self.try_populate_init_table()

    def try_populate_init_table(self):
        if (
            self.env_box.relative_to_curr.isChecked()
            and self.env_box.var_table.selected
        ):
            self.update_init_table()

    def handle_pv_added(self):
        if self.env_box.relative_to_curr.isChecked():
            self.set_vrange()

    def handle_var_config(self, vname):
        env = self.create_env()
        curr = env._get_variables([vname])[vname]

        # Get the hard limit
        try:
            bounds = self.var_hard_limit[vname]
        except KeyError:
            try:
                bounds = env._get_bounds([vname])[vname]
            except BadgerEnvVarError as e:
                msg = str(e)
                bounds = eval(msg.split(": ")[1])

        # Get the option
        try:
            option = self.ratio_var_ranges[vname]
        except KeyError:
            option = self.limit_option

        configs = {
            "current_value": curr,
            "lower_bound": bounds[0],
            "upper_bound": bounds[1],
            **option,
        }

        BadgerIndividualLimitVariableRangeDialog(
            self,
            vname,
            self.set_ind_vrange,
            configs,
        ).exec_()

    def add_constraint(self, name=None, relation=0, threshold=0, critical=False):
        if self.configs is None:
            return

        options = self.configs["observations"]
        self.env_box.con_table.add_constraint(
            options,
            name,
            relation,
            threshold,
            critical,
        )

    def add_state(self, name=None):
        if self.configs is None:
            return

        var_names = [next(iter(d)) for d in self.configs["variables"]]
        options = self.configs["observations"] + var_names
        item = QListWidgetItem(self.env_box.list_obs)
        sta_item = state_item(
            options,
            lambda: self.env_box.list_obs.takeItem(self.env_box.list_obs.row(item)),
            name,
        )
        item.setSizeHint(sta_item.sizeHint())
        self.env_box.list_obs.addItem(item)
        self.env_box.list_obs.setItemWidget(item, sta_item)
        self.env_box.fit_content()

    def _compose_vocs(self) -> (VOCS, list[str]):
        # Compose the VOCS settings
        variables = self.env_box.var_table.export_variables()
        objectives = self.env_box.obj_table.export_objectives()

        constraints = {}
        critical_constraints = []
        for (
            con_name,
            relation,
            threshold,
            critical,
            _,
        ) in self.env_box.con_table.export_constraints():
            constraints[con_name] = [relation, threshold]
            if critical:
                critical_constraints.append(con_name)

        observables = []
        for i in range(self.env_box.list_obs.count()):
            item = self.env_box.list_obs.item(i)
            item_widget = self.env_box.list_obs.itemWidget(item)
            obs_name = item_widget.cb_sta.currentText()
            observables.append(obs_name)

        try:
            vocs = VOCS(
                variables=variables,
                objectives=objectives,
                constraints=constraints,
                constants={},
                observables=observables,
            )
        except ValidationError as e:
            raise BadgerRoutineError(
                f"\n\nVOCS validation failed: {format_validation_error(e)}"
            ) from e

        return vocs, critical_constraints

    def _compose_routine(self) -> Routine:
        # Compose the routine
        # Metadata
        name = self.edit_save.text() or self.edit_save.placeholderText()
        description = self.edit_descr.toPlainText()

        # General sanity checks
        if self.generator_box.cb.currentIndex() == -1:
            raise BadgerRoutineError("no generator selected")
        if self.env_box.cb.currentIndex() == -1:
            raise BadgerRoutineError("no environment selected")

        # Generator
        generator_name = self.generators[self.generator_box.cb.currentIndex()]
        env_name = self.envs[self.env_box.cb.currentIndex()]
        generator_params = load_config(self.generator_box.edit.toPlainText())
        if generator_name in all_generator_names["bo"]:
            # Patch the BO generators to make sure use_low_noise_prior is False
            if "gp_constructor" not in generator_params:
                generator_params["gp_constructor"] = {
                    "name": "standard",  # have to add name too for pydantic validation
                    "use_low_noise_prior": False,
                }
            # or else we use whatever specified by the users

            if generator_name != "mobo":
                # Patch the BO generators to turn on TuRBO by default
                if "turbo_controller" not in generator_params:
                    generator_params["turbo_controller"] = "optimize"

                # TODO: remove this patch when Xopt reset API works
                # Nullify a few properties in turbo that can cause issues
                turbo_config = generator_params["turbo_controller"]
                if type(turbo_config) is dict:
                    if turbo_config["name"] == "optimize":
                        turbo_config["center_x"] = None
                        turbo_config["best_value"] = None
                    elif turbo_config["name"] == "safety":
                        turbo_config["center_x"] = None

        # Environment
        env_params = load_config(self.env_box.edit.toPlainText())

        # VOCS
        vocs, critical_constraints = self._compose_vocs()
        if not vocs.variables:
            raise BadgerRoutineError("no variables selected")
        if not vocs.objectives:
            raise BadgerRoutineError("no objectives selected")

        # Initial points
        init_points_df = pd.DataFrame.from_dict(
            get_table_content_as_dict(self.env_box.init_table)
        )
        init_points_df = init_points_df.replace("", pd.NA)
        init_points_df = init_points_df.dropna(subset=init_points_df.columns, how="all")
        if init_points_df.empty:
            raise BadgerRoutineError(
                "No initial points provided. Please add at least one initial point"
            )
        contains_na = init_points_df.isna().any().any()
        if contains_na:
            raise BadgerRoutineError(
                "Initial points are not valid, please fill in the missing values"
            )

        # Script that generates generator params
        if self.generator_box.check_use_script.isChecked():
            script = self.script
        else:
            script = None

        # Relative to current params
        if self.env_box.relative_to_curr.isChecked():
            relative_to_current = True
            vrange_limit_options = self.ratio_var_ranges
            initial_point_actions = self.init_table_actions
        else:
            relative_to_current = False
            vrange_limit_options = None
            initial_point_actions = None

        # Save hard limits no matter relative to current or not
        vrange_hard_limit = self.var_hard_limit

        try:
            generator = get_generator_dynamic(generator_name)(
                vocs=vocs, **generator_params
            )
        except ValidationError as e:
            raise BadgerRoutineError(
                f"\n\nAlgorithm validation failed: {format_validation_error(e)}"
            ) from e

        with warnings.catch_warnings(record=True) as caught_warnings:
            routine = Routine(
                # Metadata
                badger_version=get_badger_version(),
                xopt_version=get_xopt_version(),
                # Xopt part
                vocs=vocs,
                generator=generator,
                # Badger part
                name=name,
                description=description,
                environment={"name": env_name} | env_params,
                initial_points=init_points_df.astype("double"),
                critical_constraint_names=critical_constraints,
                tags=None,
                script=script,
                relative_to_current=relative_to_current,
                vrange_limit_options=vrange_limit_options,
                vrange_hard_limit=vrange_hard_limit,
                initial_point_actions=initial_point_actions,
                additional_variables=self.env_box.var_table.addtl_vars,
                formulas=self.env_box.obj_table.formulas,
            )

            # Check if any user warnings were caught
            for warning in caught_warnings:
                if issubclass(warning.category, UserWarning):
                    QMessageBox.warning(
                        self,
                        "Warning!",
                        f"Warning: {warning.message}",
                    )
                else:
                    print(f"Caught warning: {warning.message}")

            return routine

    def review(self):
        try:
            routine = self._compose_routine()
        except Exception:
            return QMessageBox.critical(
                self, "Invalid routine!", traceback.format_exc()
            )

        dlg = BadgerReviewDialog(self, routine)
        dlg.exec()

    def update_description(self):
        routine = self.routine
        routine.description = self.edit_descr.toPlainText()
        try:
            update_run(routine)
            # Notify routine list to update
            self.sig_updated.emit(routine.name, routine.description)
            QMessageBox.information(
                self,
                "Update success!",
                f"Routine {self.routine.name} description was updated!",
            )
        except Exception:
            return QMessageBox.critical(self, "Update failed!", traceback.format_exc())



from pandas import DataFrame
from PyQt5.QtWidgets import QApplication, QTableWidget, QTableWidgetItem
from PyQt5.QtCore import Qt


stylesheet = """
    QTableWidget
    {
        alternate-background-color: #262E38;
    }
    QTableWidget::item::selected
    {
        background-color: #B3E5FC;
        color: #000000;
    }
"""

stylesheet_data = """
    QTableWidget
    {
        margin: 0px 8px 8px 8px;
        alternate-background-color: #262E38;
    }
    QTableWidget::item::selected
    {
        background-color: #B3E5FC;
        color: #000000;
    }
"""


# https://stackoverflow.com/questions/60715462/how-to-copy-and-paste-multiple-cells-in-qtablewidget-in-pyqt5
class TableWithCopy(QTableWidget):
    """
    this class extends QTableWidget
    * supports copying multiple cell's text onto the clipboard
    * formatted specifically to work with multiple-cell paste into programs
      like google sheets, excel, or numbers
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def keyPressEvent(self, event):
        super().keyPressEvent(event)
        if event.key() == Qt.Key.Key_C and (
            event.modifiers() & Qt.KeyboardModifier.ControlModifier
        ):
            copied_cells = sorted(self.selectedIndexes())

            copy_text = ""
            max_column = copied_cells[-1].column()
            for c in copied_cells:
                copy_text += self.item(c.row(), c.column()).text()
                if c.column() == max_column:
                    copy_text += "\n"
                else:
                    copy_text += "\t"

            QApplication.clipboard().setText(copy_text)

    def set_uneditable(self):
        self.setEditTriggers(QTableWidget.NoEditTriggers)

    def set_editable(self):
        self.setEditTriggers(QTableWidget.DoubleClicked)


def format_value(v):
    try:
        f = f"{v:.4g}"
    except (ValueError, TypeError):
        f = str(v)
    return f


def update_table(table, data=None, vocs=None):
    table.setRowCount(0)
    table.horizontalHeader().setVisible(False)

    if data is None:
        return table

    if vocs is None:
        raise ValueError("vocs must be provided to update the table")

    columns = vocs.output_names + vocs.variable_names
    _data = data[columns]

    m, n = _data.shape
    table.setRowCount(m)
    table.setColumnCount(n)
    for i in range(m):
        for j in range(n):
            v = _data.iloc[i, j]
            if isinstance(v, str):
                table.setItem(i, j, QTableWidgetItem(v))
            else:
                table.setItem(i, j, QTableWidgetItem(format_value(v)))
    table.setHorizontalHeaderLabels(list(_data.columns))
    table.setVerticalHeaderLabels(
        list(map(str, _data.index))
    )  # row index starts from 0
    table.horizontalHeader().setVisible(True)

    return table


def reset_table(table, header):
    table.setRowCount(0)
    # Need to set col num or the old col num will be used for new data,
    # resulting in potential incomplete table
    table.setColumnCount(len(header))
    table.horizontalHeader().setVisible(False)
    table.setHorizontalHeaderLabels(header)
    table.horizontalHeader().setVisible(True)

    return table


def add_row(table, row):
    r = table.rowCount()
    table.insertRow(r)
    for i, v in enumerate(row):
        table.setItem(r, i, QTableWidgetItem(format_value(v)))
    table.setVerticalHeaderItem(r, QTableWidgetItem(str(r)))

    return table


def data_table(data=None):
    table = TableWithCopy()
    table.setAlternatingRowColors(True)
    table.setStyleSheet(stylesheet_data)
    # table.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
    return update_table(table, data)


def init_data_table(variable_names=None):
    table = TableWithCopy()
    table.setAlternatingRowColors(True)
    table.setStyleSheet(stylesheet)
    # table.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)

    table.setRowCount(10)
    if variable_names is None:
        return table

    table.setColumnCount(len(variable_names))
    table.horizontalHeader().setVisible(False)
    table.setHorizontalHeaderLabels(variable_names)
    table.horizontalHeader().setVisible(True)

    return table


def get_horizontal_header_as_list(table):
    header = table.horizontalHeader()
    header_labels = [
        header.model().headerData(i, header.orientation())
        for i in range(header.count())
    ]
    return header_labels


def get_table_content_as_dict(table):
    table_content = {}
    header_labels = get_horizontal_header_as_list(table)

    for col in range(table.columnCount()):
        column_name = header_labels[col]
        column_values = []

        for row in range(table.rowCount()):
            item = table.item(row, col)
            if item is not None:
                column_values.append(item.text())
            else:
                column_values.append("")

        table_content[column_name] = column_values

    return table_content


def update_init_data_table(table, variable_names):
    current_init_data = get_table_content_as_dict(table)

    table.setColumnCount(len(variable_names))
    table.horizontalHeader().setVisible(False)
    table.setHorizontalHeaderLabels(variable_names)
    table.horizontalHeader().setVisible(True)

    for col, name in enumerate(variable_names):
        if name in current_init_data:
            for row in range(table.rowCount()):
                table.setItem(row, col, QTableWidgetItem(current_init_data[name][row]))
        else:
            for row in range(table.rowCount()):
                table.setItem(row, col, QTableWidgetItem(""))


def set_init_data_table(table, data: DataFrame):
    variable_names = get_horizontal_header_as_list(table)
    try:
        data_dict = data.to_dict("list")
    except AttributeError:
        data_dict = None

    # Clear the table
    for col, name in enumerate(variable_names):
        for row in range(table.rowCount()):
            table.setItem(row, col, QTableWidgetItem(""))

    if data_dict is None:
        return

    # Fill the table
    for col, name in enumerate(variable_names):
        for row in range(len(data_dict[name])):
            table.setItem(
                row, col, QTableWidgetItem(format_value(data_dict[name][row]))
            )



import logging

import pandas as pd
import yaml

from badger.utils import yprint

logger = logging.getLogger(__name__)


def show_routine(args):
    print(
        "This command is deprecated.\n"
        "Please use 'badger -g' to launch the Badger GUI "
        "and manage routines/runs."
    )
    return

    try:
        from badger.db import load_routine, list_routine
        from badger.actions.run import run_n_archive
    except Exception as e:
        logger.error(e)
        return

    # List routines
    if args.routine_id is None:
        routines = list_routine()[1]
        if routines:
            yprint(routines)
        else:
            print("No routine has been saved yet")
        return

    try:
        routine, _ = load_routine(args.routine_id)
        if routine is None:
            print(f"Routine {args.routine_id} not found")
            return
    except Exception as e:
        print(e)
        return

    # Print the routine
    if not args.run:
        info = yaml.safe_load(routine.yaml())
        output = {}
        output["name"] = info["name"]
        output["environment"] = info["environment"]
        output["algorithm"] = info["generator"]
        output["vocs"] = info["vocs"]
        output["initial_points"] = pd.DataFrame(info["initial_points"]).to_dict("list")
        output["critical_constraint_names"] = info["critical_constraint_names"]
        output["tags"] = info["tags"]
        output["script"] = info["script"]

        yprint(output)
        return

    run_n_archive(routine, args.yes, False, args.verbose)



from copy import deepcopy

from PyQt5.QtWidgets import QDialog, QWidget, QHBoxLayout, QStackedWidget
from PyQt5.QtWidgets import QVBoxLayout, QSpinBox, QPushButton
from PyQt5.QtWidgets import QGroupBox, QLabel, QComboBox, QStyledItemDelegate
from PyQt5.QtCore import Qt
from badger.gui.default.components.robust_spinbox import RobustSpinBox


class BadgerAddRandomDialog(QDialog):
    def __init__(self, parent, add_points, save_config, configs=None):
        super().__init__(parent)

        self.add_points = add_points
        self.save_config = save_config
        self.configs = deepcopy(configs)
        if configs is None:
            self.configs = {
                "method": 0,
                "n_points": 3,
                "fraction": 0.1,
            }

        self.init_ui()
        self.config_logic()

    def init_ui(self):
        self.setWindowTitle("Add random points")
        self.setMinimumWidth(360)

        vbox = QVBoxLayout(self)

        # Action bar
        action_bar = QWidget()
        hbox = QHBoxLayout(action_bar)
        hbox.setContentsMargins(0, 0, 0, 0)

        lbl = QLabel("Add points in")
        # lbl.setFixedWidth(64)
        self.cb = cb = QComboBox()
        cb.setItemDelegate(QStyledItemDelegate())
        cb.addItems(
            [
                "region around current value",
                # 'lhs',
            ]
        )
        cb.setCurrentIndex(self.configs["method"])

        hbox.addWidget(lbl)
        hbox.addWidget(cb, 1)

        # Config group
        group_config = QGroupBox("Parameters")
        vbox_config = QVBoxLayout(group_config)

        self.stacks = stacks = QStackedWidget()
        # Xopt config
        xopt_config = QWidget()
        vbox_xopt = QVBoxLayout(xopt_config)
        vbox_xopt.setContentsMargins(0, 0, 0, 0)
        hbox_xopt = QHBoxLayout()
        hbox_xopt.setContentsMargins(0, 0, 0, 0)
        lbl = QLabel("Number of points")
        lbl.setFixedWidth(128)
        self.sb_np = sb_np = QSpinBox()
        sb_np.setMinimum(1)
        sb_np.setMaximum(100000)
        sb_np.setValue(self.configs["n_points"])
        sb_np.setSingleStep(1)
        hbox_xopt.addWidget(lbl)
        hbox_xopt.addWidget(sb_np, 1)
        vbox_xopt.addLayout(hbox_xopt)

        hbox_frac = QHBoxLayout()
        hbox_frac.setContentsMargins(0, 0, 0, 0)
        lbl_frac = QLabel("Fraction")
        lbl_frac.setFixedWidth(128)
        hbox_frac.addWidget(lbl_frac)
        self.sb_frac = sb_frac = RobustSpinBox(
            default_value=0.1, lower_bound=0.0, upper_bound=1.0, decimals=2
        )
        sb_frac.setValue(self.configs["fraction"])
        hbox_frac.addWidget(sb_frac, 1)
        vbox_xopt.addLayout(hbox_frac)

        stacks.addWidget(xopt_config)

        stacks.setCurrentIndex(self.configs["method"])
        vbox_config.addWidget(stacks)
        vbox_config.addStretch(1)

        # Button set
        button_set = QWidget()
        hbox_set = QHBoxLayout(button_set)
        hbox_set.setContentsMargins(0, 0, 0, 0)
        self.btn_cancel = btn_cancel = QPushButton("Cancel")
        self.btn_add = btn_add = QPushButton("Add")
        btn_cancel.setFixedSize(96, 24)
        btn_add.setFixedSize(96, 24)
        hbox_set.addStretch()
        hbox_set.addWidget(btn_cancel)
        hbox_set.addWidget(btn_add)

        vbox.addWidget(action_bar)
        vbox.addWidget(group_config, 1)
        vbox.addWidget(button_set)

    def config_logic(self):
        self.cb.currentIndexChanged.connect(self.method_changed)
        self.btn_cancel.clicked.connect(self.close)
        self.btn_add.clicked.connect(self.add)
        self.sb_np.valueChanged.connect(self.n_points_changed)
        self.sb_frac.valueChanged.connect(self.frac_changed)

    def n_points_changed(self, n_points):
        self.configs["n_points"] = n_points

    def frac_changed(self, frac):
        self.configs["fraction"] = frac

    def add(self):
        self.save_config(self.configs)
        self.add_points()
        self.close()

    def method_changed(self, i):
        self.stacks.setCurrentIndex(i)

        # Update configs
        self.configs["method"] = i

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Return or event.key() == Qt.Key_Enter:
            self.add()
        elif event.key() == Qt.Key_Escape:
            self.close()



import gc
import traceback
from importlib import resources

from pandas import DataFrame
from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtGui import QIcon, QKeySequence
from PyQt5.QtWidgets import (
    QMessageBox,
    QShortcut,
    QSplitter,
    QVBoxLayout,
    QWidget,
    QLabel,
)

from badger.archive import (
    delete_run,
    get_base_run_filename,
    load_run,
    get_runs,
    save_tmp_run,
)
from badger.gui.default.components.data_table import (
    add_row,
    data_table,
    reset_table,
    update_table,
)
from badger.gui.acr.components.history_navigator import HistoryNavigator
from badger.gui.acr.components.routine_editor import BadgerRoutineEditor
from badger.gui.default.components.run_monitor import BadgerOptMonitor
from badger.gui.acr.components.status_bar import BadgerStatusBar
from badger.gui.acr.components.action_bar import BadgerActionBar
from badger.utils import get_header
from badger.settings import init_settings

# from PyQt5.QtGui import QBrush, QColor
from badger.gui.default.windows.message_dialog import BadgerScrollableMessageBox
from badger.gui.default.utils import ModalOverlay

import logging

logger = logging.getLogger(__name__)

stylesheet = """
QPushButton:hover:pressed
{
    background-color: #C7737B;
}
QPushButton:hover
{
    background-color: #BF616A;
}
QPushButton
{
    background-color: #A9444E;
}
"""


class BadgerHomePage(QWidget):
    sig_routine_activated = pyqtSignal(bool)
    sig_routine_invalid = pyqtSignal()

    def __init__(self, process_manager=None):
        super().__init__()

        self.mode = "regular"  # home page mode
        self.splitter_state = None  # store the run splitter state
        self.process_manager = process_manager
        self.current_routine = None  # current routine
        self.go_run_failed = False  # flag to indicate go_run failed
        self.init_ui()
        self.config_logic()

        self.load_all_runs()
        self.init_home_page()

    def init_ui(self):
        self.config_singleton = init_settings()
        icon_ref = resources.files(__package__) / "../images/add.png"

        with resources.as_file(icon_ref) as icon_path:
            self.icon_add = QIcon(str(icon_path))
        icon_ref = resources.files(__package__) / "../images/import.png"
        with resources.as_file(icon_ref) as icon_path:
            self.icon_import = QIcon(str(icon_path))
        icon_ref = resources.files(__package__) / "../images/export.png"
        with resources.as_file(icon_ref) as icon_path:
            self.icon_export = QIcon(str(icon_path))

        # Set up the layout
        vbox = QVBoxLayout(self)
        vbox.setContentsMargins(0, 0, 0, 0)
        vbox.setSpacing(0)

        # History run browser
        self.history_browser = history_browser = HistoryNavigator()
        history_browser.setFixedWidth(360)

        # Splitter
        splitter = QSplitter(Qt.Horizontal)
        splitter.setStretchFactor(0, 0)
        splitter.setStretchFactor(1, 1)

        vbox.addWidget(splitter, 1)

        # Monitor panel
        panel_monitor = QWidget()
        vbox_monitor = QVBoxLayout(panel_monitor)
        vbox_monitor.setContentsMargins(8, 0, 8, 0)

        # Run monitor
        self.run_view = run_view = QWidget()  # for consistent bg
        vbox_monitor.addWidget(run_view)
        vbox_run_view = QVBoxLayout(run_view)
        vbox_run_view.setContentsMargins(0, 0, 0, 10)
        self.run_monitor = run_monitor = BadgerOptMonitor(self.process_manager)
        vbox_run_view.addWidget(run_monitor)

        # Data table
        panel_table = QWidget()
        panel_table.setMinimumHeight(180)
        vbox_table = QVBoxLayout(panel_table)
        vbox_table.setContentsMargins(0, 0, 0, 0)
        title_label = QLabel("Run Data")
        title_label.setStyleSheet(
            """
            background-color: #455364;
            font-weight: bold;
            padding: 4px;
        """
        )
        title_label.setAlignment(Qt.AlignCenter)  # Center-align the title
        vbox_table.addWidget(title_label, 0)
        self.run_table = run_table = data_table()
        run_table.set_uneditable()  # should not be editable
        vbox_table.addWidget(run_table, 1)

        # Routine view
        self.routine_view = routine_view = QWidget()  # for consistent bg
        routine_view.setMinimumWidth(640)
        vbox_routine_view = QVBoxLayout(routine_view)
        vbox_routine_view.setContentsMargins(0, 0, 0, 10)
        self.routine_editor = routine_editor = BadgerRoutineEditor()
        vbox_routine_view.addWidget(routine_editor)

        # Add action bar
        self.run_action_bar = run_action_bar = BadgerActionBar()

        # Run panel (routine editor + run monitor + data table + action bar)
        panel_run = QWidget()
        vbox_run = QVBoxLayout(panel_run)
        vbox_run.setContentsMargins(0, 0, 0, 0)
        vbox_run.setSpacing(0)

        splitter_run = QSplitter(Qt.Horizontal)
        splitter_run.setStretchFactor(0, 1)
        splitter_run.setStretchFactor(1, 1)
        vbox_run.addWidget(splitter_run, 1)

        splitter_data = QSplitter(Qt.Vertical)
        splitter_data.setStretchFactor(0, 1)
        splitter_data.setStretchFactor(1, 0)
        splitter_data.addWidget(panel_monitor)
        splitter_data.addWidget(panel_table)

        splitter_run.addWidget(routine_view)
        splitter_run.addWidget(splitter_data)

        vbox_run.addWidget(run_action_bar, 0)

        # Add panels to splitter
        splitter.addWidget(history_browser)
        splitter.addWidget(panel_run)

        # Set initial sizes (left fixed, middle and right equal)
        splitter.setSizes([1, 1])
        splitter_run.setSizes([1, 1])
        splitter_data.setSizes([800, 180])

        self.status_bar = status_bar = BadgerStatusBar()
        status_bar.set_summary("Badger is ready!")
        vbox.addWidget(status_bar)

    def config_logic(self):
        self.colors = ["c", "g", "m", "y", "b", "r", "w"]
        self.symbols = ["o", "t", "t1", "s", "p", "h", "d"]

        self.run_table.cellClicked.connect(self.solution_selected)
        self.run_table.itemSelectionChanged.connect(self.table_selection_changed)

        self.history_browser.tree_widget.itemSelectionChanged.connect(self.go_run)

        self.routine_editor.routine_page.sig_load_template.connect(self.update_status)
        self.routine_editor.routine_page.sig_save_template.connect(self.update_status)

        self.run_monitor.sig_inspect.connect(self.inspect_solution)
        self.run_monitor.sig_lock.connect(self.toggle_lock)
        self.run_monitor.sig_new_run.connect(self.new_run)
        self.run_monitor.sig_run_started.connect(self.uncover_page)
        self.run_monitor.sig_run_name.connect(self.run_name)
        self.run_monitor.sig_status.connect(self.update_status)
        self.run_monitor.sig_progress.connect(self.progress)
        self.run_monitor.sig_del.connect(self.delete_run)
        self.run_monitor.sig_stop_run.connect(self.cover_page)
        self.run_monitor.sig_run_started.connect(self.run_action_bar.run_start)
        self.run_monitor.sig_routine_finished.connect(
            self.run_action_bar.routine_finished
        )
        self.run_monitor.sig_lock_action.connect(self.run_action_bar.lock)
        self.run_monitor.sig_toggle_reset.connect(self.run_action_bar.toggle_reset)
        self.run_monitor.sig_toggle_run.connect(self.run_action_bar.toggle_run)
        self.run_monitor.sig_toggle_other.connect(self.run_action_bar.toggle_other)
        self.run_monitor.sig_env_ready.connect(self.run_action_bar.env_ready)

        self.run_action_bar.sig_start.connect(self.start_run)
        self.run_action_bar.sig_start_until.connect(self.start_run_until)
        self.run_action_bar.sig_stop.connect(self.run_monitor.stop)
        self.run_action_bar.sig_delete_run.connect(self.run_monitor.delete_run)
        self.run_action_bar.sig_logbook.connect(self.run_monitor.logbook)
        self.run_action_bar.sig_reset_env.connect(self.run_monitor.reset_env)
        self.run_action_bar.sig_jump_to_optimal.connect(
            self.run_monitor.jump_to_optimal
        )
        self.run_action_bar.sig_dial_in.connect(self.run_monitor.set_vars)
        self.run_action_bar.sig_ctrl.connect(self.run_monitor.ctrl_routine)
        self.run_action_bar.sig_open_extensions_palette.connect(
            self.run_monitor.open_extensions_palette
        )

        self.sig_routine_invalid.connect(self.run_action_bar.routine_invalid)

        # Assign shortcuts
        self.shortcut_go_search = QShortcut(QKeySequence("Ctrl+L"), self)
        self.shortcut_go_search.activated.connect(self.go_search)

    def go_search(self):
        self.sbar.setFocus()

    def load_all_runs(self):
        runs = get_runs()
        self.history_browser.updateItems(runs)

    def init_home_page(self):
        # Load the default generator
        self.routine_editor.routine_page.generator_box.cb.setCurrentIndex(0)

    def go_run(self, i: int = None):
        gc.collect()

        # if self.cb_history.itemText(0) == "Optimization in progress...":
        #     return

        if i == -1:
            update_table(self.run_table)
            try:
                self.current_routine.data = None  # reset the data
            except AttributeError:  # current routine is None
                pass
            self.run_monitor.init_plots(self.current_routine)
            if not self.current_routine:
                self.routine_editor.set_routine(None)
                self.status_bar.set_summary("No active routine")
            else:
                self.status_bar.set_summary(
                    f"Current routine: {self.current_routine.name}"
                )
            return

        run_filename = get_base_run_filename(self.history_browser.currentText())
        try:
            routine = load_run(run_filename)
            self.run_monitor.routine_filename = run_filename
        except IndexError:
            return
        except Exception as e:  # failed to load the run
            details = traceback.format_exc()
            dialog = BadgerScrollableMessageBox(
                title="Error!", text=str(e), parent=self
            )
            dialog.setIcon(QMessageBox.Critical)
            dialog.setDetailedText(details)
            dialog.exec_()
            self.go_run_failed = True

            # Show info in the nav bar
            # red_brush = QBrush(QColor(255, 0, 0))  # red color
            # self.cb_history.changeCurrentItem(
            #     f"{run_filename} (failed to load)",
            #     # color=red_brush)
            #     color=None,
            # )

            return

        self.current_routine = routine  # update the current routine
        update_table(self.run_table, routine.sorted_data, routine.vocs)
        self.run_monitor.init_plots(routine, run_filename)
        self.routine_editor.set_routine(routine, silent=True)
        self.status_bar.set_summary(f"Current routine: {self.current_routine.name}")

        self.run_monitor.update_analysis_extensions()

    def inspect_solution(self, idx):
        self.run_table.selectRow(idx)

    def solution_selected(self, r, c):
        self.run_monitor.jump_to_solution(r)

    def table_selection_changed(self):
        indices = self.run_table.selectedIndexes()
        if len(indices) == 1:  # let other method handles it
            return

        row = -1
        for index in indices:
            _row = index.row()
            if _row == row:
                continue

            if row == -1:
                row = _row
                continue

            return

        if row == -1:
            return

        self.run_monitor.jump_to_solution(row)

    def toggle_lock(self, lock, lock_tab=1):
        if lock:
            self.history_browser.setDisabled(True)
        else:
            self.history_browser.setDisabled(False)

            self.uncover_page()

    def prepare_run(self):
        try:
            routine = self.routine_editor.routine_page._compose_routine()
        except Exception as e:
            self.sig_routine_invalid.emit()
            raise e

        self.current_routine = routine

        # Save routine as a temp file
        # since routine runner subprocess needs to load routine from file
        run_filename = save_tmp_run(routine)
        self.run_monitor.routine_filename = run_filename

        # Tell monitor to start the run
        self.run_monitor.init_plots(routine)

    def start_run(self):
        self.prepare_run()
        self.run_monitor.start()

    def start_run_until(self):
        self.prepare_run()
        self.run_monitor.start_until()

    def new_run(self):
        self.cover_page()

        # self.cb_history.insertItem(0, "Optimization in progress...")
        # self.cb_history.setCurrentIndex(0)

        header = get_header(self.current_routine)
        reset_table(self.run_table, header)

    def run_name(self, name):
        runs = get_runs()
        self.history_browser.updateItems(runs)
        self.history_browser._selectItemByRun(name)

    def update_status(self, info):
        self.status_bar.set_summary(info)

    def progress(self, solution: DataFrame):
        vocs = self.current_routine.vocs
        vars = list(solution[vocs.variable_names].to_numpy()[0])
        objs = list(solution[vocs.objective_names].to_numpy()[0])
        cons = list(solution[vocs.constraint_names].to_numpy()[0])
        stas = list(solution[vocs.observable_names].to_numpy()[0])
        add_row(self.run_table, objs + cons + vars + stas)

    def delete_run(self):
        run_name = get_base_run_filename(self.history_browser.currentText())

        reply = QMessageBox.question(
            self,
            "Delete run",
            f"Are you sure you want to delete run {run_name}?",
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No,
        )
        if reply != QMessageBox.Yes:
            return

        delete_run(run_name)
        runs = get_runs()
        self.history_browser.tree_widget.blockSignals(True)
        self.history_browser.updateItems(runs)
        self.history_browser.tree_widget.blockSignals(False)
        self.go_run(-1)

    def cover_page(self):
        return  # disable overlay for now

        try:
            self.overlay
        except AttributeError:
            # Set parent to the main window
            try:
                main_window = self.parent().parent()
            except AttributeError:  # in test mode
                return
            self.overlay = ModalOverlay(main_window)
        self.overlay.show()

    def uncover_page(self):
        return  # disable overlay for now

        try:
            self.overlay.hide()
        except AttributeError:  # in test mode
            pass



from abc import ABC
from logging import warning
from typing import ClassVar, Dict, final, List, Optional

from pydantic import BaseModel, ConfigDict, Field, SerializeAsAny

from badger.errors import (
    BadgerEnvVarError,
    BadgerNoInterfaceError,
)
from badger.interface import Interface


def validate_setpoints(func):
    def validate(cls, variable_inputs: Dict[str, float]):
        _bounds = cls._get_bounds(list(variable_inputs.keys()))
        for name, value in variable_inputs.items():
            lower = _bounds[name][0]
            upper = _bounds[name][1]

            if value > upper or value < lower:
                raise BadgerEnvVarError(
                    f"Input point for {name} is outside "
                    + f"its bounds {_bounds[name]}"
                )

        return func(cls, variable_inputs)

    return validate


class Environment(BaseModel, ABC):
    model_config = ConfigDict(
        validate_assignment=True, use_enum_values=True, arbitrary_types_allowed=True
    )

    # Class variables
    name: ClassVar[str] = Field(description="environment name")
    variables: ClassVar[Dict[str, List]]  # bounds list could be empty for var
    observables: ClassVar[List[str]]

    # Interface
    interface: Optional[SerializeAsAny[Interface]] = None
    # Put all other env params here
    # params: float = Field(..., description='Example env parameter')

    ############################################################
    # Optional methods to inherit
    ############################################################

    def get_variables(self, variable_names: List[str]) -> Dict:
        if not self.interface:
            raise BadgerNoInterfaceError

        return self.interface.get_values(variable_names)

    def set_variables(self, variable_inputs: Dict[str, float]):
        if not self.interface:
            raise BadgerNoInterfaceError

        return self.interface.set_values(variable_inputs)

    def get_observables(self, observable_names: List[str]) -> Dict:
        if not self.interface:
            raise BadgerNoInterfaceError

        return self.interface.get_values(observable_names)

    def reset_environment(self):
        if self.interface:
            return self.interface.reset_interface()

    def get_bounds(self, variable_names: List[str]) -> Dict[str, List[float]]:
        raise NotImplementedError

    # Actions to preform after changing vars and before reading vars/obj
    def variables_changed(self, variables_input: Dict[str, float]):
        pass

    # Get current system states
    # If return is not None, the states would be saved at the start of each run
    # Should return a dict if not None
    def get_system_states(self) -> Optional[Dict]:
        return {}

    # Shortcuts for getting/setting single variable
    def get_variable(self, variable_name):
        return self.get_variables([variable_name])[variable_name]

    def set_variable(self, variable_name, variable_value):
        return self.set_variables({variable_name: variable_value})

    def get_bound(self, variable_name):
        return self.get_bounds([variable_name])[variable_name]

    def search(self, keyword: str, callback: callable):
        return None

    ############################################################
    # Expert level of customization
    ############################################################

    # @model_serializer
    # def ser_model(self, **kwargs) -> Dict[str, Any]:
    #     default_dict = super().model_dump(**kwargs)
    #     default_dict["name"] = self.name

    #     return default_dict

    ############################################################
    # Should never be overridden
    ############################################################

    @final
    @property
    def variable_names(self):
        return [k for k in self.variables]

    # Optimizer will only call this method to get variable values
    @final
    def _get_variables(self, variable_names: List[str]) -> Dict:
        # We'll let the users handle the case when the variable is not defined
        variable_outputs = self.get_variables(variable_names)

        return variable_outputs

    # Optimizer will only call this method to set variable values
    @final
    @validate_setpoints
    def _set_variables(self, variable_inputs: Dict[str, float]):
        self.set_variables(variable_inputs)

    # Optimizer will only call this method to get observable values
    @final
    def _get_observables(self, observable_names: List[str]) -> Dict:
        return self.get_observables(observable_names)

    # Optimizer will only call this method to get variable bounds
    # Lazy loading -- read the bounds only when they are needed
    @final
    def _get_bounds(
        self, variable_names: Optional[List[str]] = None
    ) -> Dict[str, List[float]]:
        if variable_names is None:
            variable_names = self.variable_names

        # variable_names_unset also includes those defined but not initialized vars
        variable_names_unset = [
            name for name in variable_names if not len(self.variables.get(name, []))
        ]

        # Get bound one by one due to potential failure
        for name in variable_names_unset:
            try:
                bound = self.get_bound(name)
            except Exception as e:
                raise e

            # Ignore invalid bounds issue for now
            # if bound[1] <= bound[0]:
            #     raise BadgerEnvVarError(f"Invalid bound for {name}: {bound}")

            # TODO:The tmp vars will go into the class vars,
            # it might be better to use a different name
            self.variables.update({name: bound})

        return {k: self.variables[k] for k in variable_names}


def instantiate_env(env_class, configs, manager=None):
    # Configure interface
    # TODO: figure out the correct logic
    # It seems that the interface should be given rather than
    # initialized here
    from badger.factory import get_intf

    try:
        intf_name = configs["interface"][0]
    except KeyError:
        intf_name = None
    except Exception as e:
        warning(e)
        intf_name = None

    if intf_name is not None:
        if manager is None:
            Interface, _ = get_intf(intf_name)
            intf = Interface()
        else:
            intf = manager.Interface()
    else:
        intf = None

    env = env_class(interface=intf, **configs["params"])

    return env



from PyQt5.QtWidgets import QComboBox, QTreeWidget, QTreeWidgetItem
from PyQt5.QtCore import QModelIndex, Qt
from badger.archive import get_base_run_filename
from badger.utils import run_names_to_dict


# Modified based on the following solution
# https://stackoverflow.com/a/9672359/4263605
class HistoryNavigator(QComboBox):
    def __init__(self):
        super().__init__()
        # self.view() = QTreeWidget()
        self.setView(QTreeWidget())
        self.setModel(self.view().model())

        self.view().setHeaderHidden(True)
        self.view().setMinimumHeight(256)
        # self.view().setItemsExpandable(False)
        # self.view().setRootIsDecorated(False)
        self.runs = None  # all runs to be shown in combobox

    def showPopup(self):
        self.setRootModelIndex(QModelIndex())  # key to success!
        QComboBox.showPopup(self)

    def _firstSelectableItem(self, parent=QModelIndex()):
        """
        Internal recursive function for finding the first selectable item.
        """
        for i in range(self.model().rowCount(parent)):
            itemIndex = self.model().index(i, 0, parent)
            if self.view().itemFromIndex(itemIndex).flags() & Qt.ItemIsSelectable:
                return parent, i
            else:
                itemIndex, row = self._firstSelectableItem(itemIndex)
                if row is not None:
                    return itemIndex, row
        return parent, None

    def currentIsFirst(self):
        if not self.runs:
            return True

        run = get_base_run_filename(self.currentText())
        try:
            idx = self.runs.index(run)
            if idx == 0:
                return True
            else:
                return False
        except:  # run is empty string
            return True

    def currentIsLast(self):
        if not self.runs:
            return True

        run = get_base_run_filename(self.currentText())
        try:
            idx = self.runs.index(run)
            tot = len(self.runs)
            if idx == tot - 1:
                return True
            else:
                return False
        except:  # run is empty string
            return True

    def _firstMatchingSelectableItem(self, text, parent=QModelIndex()):
        for i in range(self.model().rowCount(parent)):
            itemIndex = self.model().index(i, 0, parent)
            item = self.view().itemFromIndex(itemIndex)
            if (item.flags() & Qt.ItemIsSelectable) and item.text(0).startswith(text):
                return parent, i
            else:
                itemIndex, row = self._firstMatchingSelectableItem(text, itemIndex)
                if row is not None:
                    return itemIndex, row
        return parent, None

    def _nextSelectableItem(self, parent=QModelIndex()):
        run_curr = get_base_run_filename(self.currentText())
        idx = self.runs.index(run_curr)
        run_next = self.runs[idx + 1]

        return self._firstMatchingSelectableItem(run_next)

    def _previousSelectableItem(self, parent=QModelIndex()):
        run_curr = get_base_run_filename(self.currentText())
        idx = self.runs.index(run_curr)
        run_prev = self.runs[idx - 1]

        return self._firstMatchingSelectableItem(run_prev)

    def _currentSelectableItem(self, parent=QModelIndex()):
        run_curr = get_base_run_filename(self.currentText())

        return self._firstMatchingSelectableItem(run_curr)

    def updateItems(self, runs=None):
        self.view().clear()

        self.runs = runs  # store the runs for nav
        if runs is None:
            return

        items = []
        flag_first_item = True
        first_items = []
        runs_dict = run_names_to_dict(runs)
        for year, dict_year in runs_dict.items():
            item_year = QTreeWidgetItem([year])
            item_year.setFlags(item_year.flags() & ~Qt.ItemIsSelectable)

            if flag_first_item:
                first_items.append(item_year)

            for month, dict_month in dict_year.items():
                item_month = QTreeWidgetItem([month])
                item_month.setFlags(item_month.flags() & ~Qt.ItemIsSelectable)

                if flag_first_item:
                    first_items.append(item_month)

                for day, list_day in dict_month.items():
                    item_day = QTreeWidgetItem([day])
                    item_day.setFlags(item_day.flags() & ~Qt.ItemIsSelectable)

                    if flag_first_item:
                        first_items.append(item_day)
                        flag_first_item = False

                    for i, file in enumerate(list_day):
                        item_file = QTreeWidgetItem([file])
                        item_day.addChild(item_file)
                    item_month.addChild(item_day)
                item_year.addChild(item_month)
            items.append(item_year)
        self.view().insertTopLevelItems(0, items)
        for item in first_items:
            item.setExpanded(True)

        parent, row = self._firstSelectableItem()  # key to success!
        if row is not None:
            self.setRootModelIndex(parent)
            self.setCurrentIndex(row)

    def selectNextItem(self):
        parent, row = self._nextSelectableItem()
        if row is not None:
            self.setRootModelIndex(parent)
            self.setCurrentIndex(row)

    def selectPreviousItem(self):
        parent, row = self._previousSelectableItem()
        if row is not None:
            self.setRootModelIndex(parent)
            self.setCurrentIndex(row)

    def changeCurrentItem(self, text, color=None):
        parent, row = self._currentSelectableItem()
        if row is not None:
            itemIndex = self.model().index(row, 0, parent)
            item = self.view().itemFromIndex(itemIndex)
            item.setText(0, text)
            if color:
                item.setForeground(0, color)



from PyQt5.QtWidgets import QVBoxLayout, QHBoxLayout, QWidget
from PyQt5.QtWidgets import QComboBox, QStyledItemDelegate, QLabel
from badger.gui.default.components.collapsible_box import CollapsibleBox


class BadgerFilterBox(CollapsibleBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.init_ui()

    def init_ui(self):
        vbox = QVBoxLayout()

        # Obj filter
        obj = QWidget()
        hbox_obj = QHBoxLayout(obj)
        hbox_obj.setContentsMargins(0, 0, 0, 0)
        lbl = QLabel("Objective")
        lbl.setFixedWidth(64)
        self.cb_obj = cb_obj = QComboBox()
        cb_obj.setItemDelegate(QStyledItemDelegate())
        cb_obj.addItems(["", "HXR", "SXR"])
        cb_obj.setCurrentIndex(-1)
        hbox_obj.addWidget(lbl)
        hbox_obj.addWidget(cb_obj, 1)
        vbox.addWidget(obj)

        # Region filter
        region = QWidget()
        hbox_reg = QHBoxLayout(region)
        hbox_reg.setContentsMargins(0, 0, 0, 0)
        lbl = QLabel("Region")
        lbl.setFixedWidth(64)
        self.cb_reg = cb_reg = QComboBox()
        cb_reg.setItemDelegate(QStyledItemDelegate())
        cb_reg.addItems(
            [
                "",
                "LI21:201, 211, 271, 278",
                "LI26:201, 301, 401, 501",
                "LI26:601, 701, 801, 901",
                "IN20:361, 371, 425, 441, 511, 525",
                "LTUH:620, 640, 660, 680",
                "LTUS:620, 640, 660, 680",
            ]
        )
        cb_reg.setCurrentIndex(-1)
        hbox_reg.addWidget(lbl)
        hbox_reg.addWidget(cb_reg, 1)
        vbox.addWidget(region)

        # Gain filter
        gain = QWidget()
        hbox_gain = QHBoxLayout(gain)
        hbox_gain.setContentsMargins(0, 0, 0, 8)
        lbl = QLabel("Gain")
        lbl.setFixedWidth(64)
        self.cb_gain = cb_gain = QComboBox()
        cb_gain.setItemDelegate(QStyledItemDelegate())
        cb_gain.addItems(
            [
                "",
                "1",
                "2",
                "4",
                "8",
            ]
        )
        cb_gain.setCurrentIndex(-1)
        hbox_gain.addWidget(lbl)
        hbox_gain.addWidget(cb_gain, 1)
        vbox.addWidget(gain)

        self.setContentLayout(vbox)
